#!/usr/bin/env python
#  XMMS2 - X Music Multiplexer System
#  Copyright (C) 2003-2006 XMMS2 Team
# 
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#                   
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  This file is a part of the XMMS2 client tutorial #3
#  Here we will learn to retrieve more complex results from
#  the result set

import xmmsclient
import os
import sys

"""
The first part of this program is
commented on in tut1.py and tut2.py
"""
xmms = xmmsclient.XMMS("tutorial3")
try:
	xmms.connect(os.getenv("XMMS_PATH"))
except IOError, detail:
	print "Connection failed:", detail
	sys.exit(1)

"""
Ok, let' do the same thing as we did in
tut2.py and retrieve the current playing
entry. We need that to get information
about the song.
"""
result = xmms.playback_current_id()
result.wait()
if result.iserror():
	print "playback current id returns error, %s" % result.get_error()
id = result.value()

"""Print the value"""
print "Currently playing id is %d" % id

"""
Something about the medialib and xmms2. All
entries that are played, put into playlists
have to be in the medialib. A song's metadata
will be added to the medialib the first time
you do "xmms2 add" or equivalent.

When we request information for an entry, it will
be requested from the medialib, not the playlist
or the playback. The playlist and playback only
know the unique id of the entry. All other 
information must be retrieved in subsequent calls.

Entry 0 is non valid. Only 1-inf is valid.
So let's check for 0 and don't ask medialib for it.
"""
if id == 0:
	print "Nothing is playing."
	sys.exit(1)

"""
And now for something about return types from
clientlib. The clientlib will always return
an XMMSResult that will eventually be filled.
It can be filled with int and string  as
base types. It can also be filled with more complex
types like lists and dicts. A dict is a key<->value
representation where key is always a string but
the value can be int or string.

When retrieving an entry from the medialib, you
get a dict as return. Let's print out some
entries from it and then traverse the dict.
"""
result = xmms.medialib_get_info(id)
result.wait()

if result.iserror():
	"""
	This can return error if the id
	is not in the medialib
	"""
	print "medialib get info returns error, %s" % result.get_error()
	sys.exit(1)

"""
We can extract entries from the dict as we would with
a normal Python dict:
"""
minfo = result.value()
try:
	val = minfo["artist"]
except KeyError:
	"""
	if we end up here it means that the key "artist" wasn't
	in the dict.

	Actually this is no disasater, it might just mean that
	we don't have an artist tag on this entry. Let's
	call it no artist for now.
	"""
	val = "No Artist"
print "artist = %s" % val

try:
	val = minfo["title"]
except KeyError:
	val = "No Title"
print "title = %s" % val

try:
	val = minfo["bitrate"]
except KeyError:
	val = 0
print "bitrate = %i" % val
	
sys.exit(0)



from waflib.TaskGen import before_method,feature
from waflib import Utils

"""
Propagate 'uselib' from local 'object' targets to avoid
having to specify the 'uselib' dependencies at place
that depends on some 'object'.
"""

@feature('cprogram')
@before_method('process_use')
def propagate_uselibs(self):
    lst = []
    def add_use(tg):
        for x in Utils.to_list(getattr(tg, 'uselib', '')):
            if x not in lst:
                lst.append(x)
        for x in Utils.to_list(getattr(tg, 'use', '')):
            add_use(self.bld.get_tgen_by_name(x))

    add_use(self)
    self.uselib = lst



# Copyright (C) 2003-2023 XMMS2 Team
#
# As sphinx doesn't detect constants in modules we have to
# generate parts of the documentation. To keep it all in the
# same place, all of the documentation sources are generated.

from __future__ import with_statement
import ast
import os
import sys
import subprocess
import re
from datetime import datetime

try:
    sys.path.remove(os.path.dirname(os.path.abspath(__file__)))
    import xmmsclient
    import xmmsclient.collections
    import xmmsclient.service
except:
    print("Set PYTHONPATH to a path containing xmmsclient.")
    raise SystemExit

def write_xmmsapi(filename):
    with open(filename, "w") as dst:
        dst.write("xmmsclient\n")
        dst.write("----------\n")
        dst.write(".. py:module:: xmmsclient\n")
        dst.write("\nFunctions\n")
        dst.write("^^^^^^^^^\n")
        dst.write(".. py:function:: %s\n" % xmmsclient.userconfdir_get.__doc__)
        dst.write("\nClasses\n")
        dst.write("^^^^^^^\n")
        dst.write(".. autoclass:: xmmsclient.XmmsSync\n")
        dst.write(".. autoclass:: xmmsclient.XmmsLoop\n")
        dst.write(".. py:class:: %s\n\n" % xmmsclient.xmmsapi.XmmsCore.__doc__.replace("Core", ""))
        entries = xmmsclient.Xmms.__dict__.items() + xmmsclient.xmmsapi.XmmsCore.__dict__.items()
        for name, func in sorted(entries):
            if type(func).__name__ in ('getset_descriptor'):
                dst.write("   .. py:attribute:: %s\n" % name)
        for name, func in sorted(entries):
            if name.startswith("_"):
                continue
            if type(func).__name__ in ('method_descriptor', 'cython_function_or_method'):
                docstring = func.__doc__.replace("XmmsApi.", "").replace("XmmsCore.", "")
                match = re.findall("-> (.+)", docstring, re.MULTILINE)
                if match:
                    docstring = re.sub(" -> (.+)", "", docstring, re.MULTILINE)
                    docstring += "\n\n\t\t:rtype: %s\n" % match[0]
                dst.write("   .. py:method:: %s\n" % docstring)
        dst.write(".. autoclass:: xmmsclient.XmmsValue\n")
        dst.write(".. autoclass:: xmmsclient.XmmsValueC2C\n")
        dst.write(".. autoclass:: xmmsclient.XmmsLoop\n")
        dst.write(".. autoclass:: xmmsclient.xmmsapi.XmmsResult\n")
        dst.write(".. autoclass:: xmmsclient.xmmsapi.XmmsVisChunk\n")
        dst.write("\nExceptions\n")
        dst.write("^^^^^^^^^^\n")
        dst.write(".. autoexception:: xmmsclient.xmmsvalue.XmmsError\n")
        dst.write(".. autoexception:: xmmsclient.xmmsapi.VisualizationError\n")
        dst.write(".. autoexception:: xmmsclient.xmmsapi.XmmsDisconnectException\n")

def write_service_clients(filename):
    with open(filename, "w") as dst:
        dst.write("xmmsclient.service\n")
        dst.write("------------------\n\n")
        dst.write("See `samples` on how to register and/or consume service clients.\n\n")
        dst.write("\nClasses\n")
        dst.write("^^^^^^^\n")
        dst.write(".. autoclass:: xmmsclient.service.XmmsServiceClient\n")
        dst.write(".. py:class:: xmmsclient.service.XmmsServiceNamespace\n\n")
        dst.write("   .. py:method:: register()\n\n")
        dst.write(".. py:class:: xmmsclient.service.%s\n" %
                  xmmsclient.service.method_arg.__init__.__doc__.replace(".__init__", "").replace("self, ", ""))
        dst.write(".. py:class:: xmmsclient.service.%s\n" %
                  xmmsclient.service.method_varg.__init__.__doc__.replace(".__init__", "").replace("self, ", ""))
        dst.write(".. py:class:: xmmsclient.service.%s\n" %
                  xmmsclient.service.service_constant.__init__.__doc__.replace(".__init__", "").replace("self, ", ""))
        dst.write(".. py:class:: xmmsclient.service.%s\n" %
                  xmmsclient.service.service_broadcast.__doc__)
        dst.write("\nDecorators\n")
        dst.write("^^^^^^^^^^\n")
        dst.write(".. py:decorator:: xmmsclient.service.%s\n" %
                  xmmsclient.service.service_method.__init__.im_func.func_doc.replace(".__init__", ""))


def write_constants(filename):
    with open(xmmsclient.consts.__file__) as src, open(filename, "w") as dst:
        class ConstsDocumenter(ast.NodeVisitor):
            def visit_ImportFrom(self, node):
                for name in node.names:
                    value = eval("xmmsclient.%s" % name.name)
                    dst.write(".. py:data:: xmmsclient.%s\n" % name.name)
                    dst.write("   :annotation: = %s\n\n" % value)
        dst.write("Constants\n")
        dst.write("---------\n")
        tree = ast.parse(src.read())
        ConstsDocumenter().visit(tree)

def write_collections(filename):
    with open(xmmsclient.collections.__file__) as src, open(filename, "w") as dst:
        class CollectionsDocumenter(ast.NodeVisitor):
            def visit_ImportFrom(self, node):
                for name in node.names:
                    if name.name == "coll_parse":
                        dst.write("\nMethods\n")
                        dst.write("^^^^^^^\n")
                        dst.write(".. autofunction:: xmmsclient.collections.coll_parse\n")
                    else:
                        dst.write(".. autoclass:: xmmsclient.collections.%s\n" % name.name)
                        dst.write("   :inherited-members:\n\n")
                        dst.write("   .. automethod:: xmmsclient.collections.%s.__init__\n" % name.name)
        dst.write("xmmsclient.collections\n")
        dst.write("----------------------\n")
        dst.write("\nClasses\n")
        dst.write("^^^^^^^\n")
        tree = ast.parse(src.read())
        CollectionsDocumenter().visit(tree)

def write_samples(filename, samples_dir):
    with open(filename, "w") as dst:
        dst.write("\n")
        dst.write("Examples\n")
        dst.write("--------\n")
        dst.write("\nBrowse and track available service clients\n")
        dst.write("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n")
        dst.write(".. literalinclude:: %s\n" % os.path.join(samples_dir, "scwatch.py"))
        dst.write("   :language: python\n")
        dst.write("   :tab-width: 2\n\n")
        dst.write("\nRegister a service client\n")
        dst.write("^^^^^^^^^^^^^^^^^^^^^^^^^\n")
        dst.write(".. literalinclude:: %s\n" % os.path.join(samples_dir, "service.py"))
        dst.write("   :language: python\n")
        dst.write("   :tab-width: 2\n")

def write_index(filename):
    with open(filename, "w") as dst:
        dst.write("Welcome to XMMS2's documentation!\n")
        dst.write("=================================\n")

        dst.write(".. toctree::\n")
        dst.write("   :maxdepth: 1\n\n")
        dst.write("   xmmsapi\n")
        dst.write("   collections\n")
        dst.write("   serviceclients\n")
        dst.write("   constants\n")
        dst.write("   samples\n")

def write_conf(filename, year, version):
    with open(filename, "w") as dst:
        dst.write("""
project = 'XMMS2'
copyright = '2016, XMMS2 Team'
version = '0.8+WiP'
master_doc = 'index'
html_theme = 'sphinx_py3doc_enhanced_theme'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary'
]

autodoc_default_flags = ['members', 'undoc-members']
default_role = 'any'
""" % {
    "year": year,
    "version": version
})

if __name__ == '__main__':
    if len(sys.argv) < 3:
        raise SystemExit("ERROR: Need an output directory argument.")

    basedir = sys.argv[1]
    script_dir = os.path.dirname(os.path.abspath(__file__))
    samples_dir = os.path.relpath(os.path.join(script_dir, "xmmsclient", "samples"), basedir)

    version = sys.argv[2]

    try:
        os.makedirs(basedir)
    except:
        pass
    write_xmmsapi(os.path.join(basedir, "xmmsapi.rst"))
    write_constants(os.path.join(basedir, "constants.rst"))
    write_collections(os.path.join(basedir, "collections.rst"))
    write_service_clients(os.path.join(basedir, "serviceclients.rst"))
    write_samples(os.path.join(basedir, "samples.rst"), samples_dir)
    write_index(os.path.join(basedir, "index.rst"))
    write_conf(os.path.join(basedir, "conf.py"), datetime.now().year, version)



#!/usr/bin/env python
#  XMMS2 - X Music Multiplexer System
#  Copyright (C) 2003-2006 XMMS2 Team
# 
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#                   
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  This file is a part of the XMMS2 client tutorial #6
#  Introducing asynchronous clients

import xmmsclient
import xmmsclient.glib
import os
import sys
import gobject

"""
Initialize the mainloop, for more information about GLib mainloop
see the GTK and PyGTK docs.
"""
ml = gobject.MainLoop(None, False)

"""
In an async client we still connect as
normal. Read up on this in earlier
tutorials if you need.
"""
xmms = xmmsclient.XMMS("tutorial6")
try:
	xmms.connect(os.getenv("XMMS_PATH"))
except IOError, detail:
	print "Connection failed:", detail
	sys.exit(1)

"""
We set this up as a callback for our current_id
method. Read the main program first before
returning here.
"""
def my_current_id(result):
	"""
	At this point the XMMSResult instance is filled with the
	answer. And we can now extract it as normal.
	"""
	print "Current id is %d" % result.value()
	ml.quit()

"""
The big difference between a sync client and an async client is that the
async client works with callbacks. When you send such an asynchronous
command, the callback set up will receive the result when it arrives.
That means you don't need to wait (XMMSResult.wait()) for results
as in synchronous operations.

In order to make xmmsclient call your callback functions we need to
put the fd of the connection into the mainloop of our program. In this
case, we're using the Glib mainloop, so We just need to import
xmmsclient.glib and do the following call to make it work.

Note that the XMMS object also has a loop() method that allows for
asynchronous communications with the server. However, this is only
useful in simple clients where you don't need to worry about GUI
mainloops and such.
"""
conn = xmmsclient.glib.GLibConnector(xmms)

"""
Let's ask for the current id in an async way instead of the sync way
as we did in tut2.
"""
xmms.playback_current_id(my_current_id)

"""
As you see we do it pretty much the same way that we did in tut2, but
instead of being able to access the current id directly (as we would
have if we where blocking) we need to wait until xmms2 calls our
my_current_id function. This will keep your GUI from hanging while
waiting for xmms2 to answer your command.

We are now all set to go. Just run the main loop and watch the magic.
"""
ml.run()



# Stock plugin configuration and build methods. These factor out the
# common tasks carried out by plugins in order to configure and build
# themselves.

from waflib.Errors import ConfigurationError
from copy import copy

def plugin(name, source=None, configure=False, build=False,
           build_replace=False, libs=[],
           tool='c', broken=False, output_prio=None):
    def stock_configure(conf):
        if broken:
            conf.msg('%s plugin' % name, 'disabled (broken)', color='RED')
            return

        if configure:
            configure(conf)

        conf.env.XMMS_PLUGINS_ENABLED.append(name)
        if output_prio:
            conf.env.XMMS_OUTPUT_PLUGINS.append((output_prio, name))

    def stock_build(bld):
        pat = tool=='c' and '*.c' or '*.cpp'
        obj = bld(
            features = '%(tool)s visibilityhidden' % dict(tool=tool),
            target = 'xmms_%s' % name,
            source = copy(source) or bld.path.ant_glob(pat),
            includes = '../../.. ../../include',
            uselib = ['glib2'] + libs,
            defines = 'G_LOG_DOMAIN="plugin/%s"' % name
        )

        if name in bld.env.XMMS_PLUGINS_BUILTIN:
            obj.defines = ["XMMS_PLUGIN_DESC_SYMBOL_NAME=XMMS_PLUGIN_DESC_" + name.upper()]
            obj.install_path = None
        else: # plugin target is shared library
            obj.features += ' %(tool)sshlib' % dict(tool=tool)
            obj.defines = ["XMMS_PLUGIN_DESC_SYMBOL_NAME=XMMS_PLUGIN_DESC"]
            obj.use = bld.env.xmms_shared_library and 'xmms2core' or ''
            obj.mac_bundle = bld.env.mac_bundle_enabled
            obj.install_path = '${PLUGINDIR}'

        if build:
            build(bld, obj)

    return stock_configure, stock_build



# encoding: utf-8
# Extend waf cython tool features.

import re
from waflib import Configure
from subprocess import Popen, STDOUT, PIPE

# Borrowed from Python 2.7, subprocess.py
class CalledProcessError(Exception):
    """This exception is raised when a process run by check_call() or
    check_output() returns a non-zero exit status.
    The exit status will be stored in the returncode attribute;
    check_output() will also store the output in the output attribute.
    """
    def __init__(self, returncode, cmd, output=None):
        self.returncode = returncode
        self.cmd = cmd
        self.output = output
    def __str__(self):
        return "Command '%s' returned non-zero exit status %d" % (self.cmd, self.returncode)

# Borrowed from Python 2.7, subprocess.py
def check_output(*popenargs, **kwargs):
    if 'stdout' in kwargs:
        raise ValueError('stdout argument not allowed, it will be overridden.')
    process = Popen(stdout=PIPE, *popenargs, **kwargs)
    output, unused_err = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        raise CalledProcessError(retcode, cmd, output=output)
    return output

cython_ver_re = re.compile('Cython version ([0-9.]+)')
def check_cython_version(self, version=None, minver=None, maxver=None):
	log_s = []
	if version:
		log_s.append("version=%r"%version)
		minver=version
		maxver=version
	else:
		if minver:
			log_s.append("minver=%r"%minver)
		if maxver:
			log_s.append("maxver=%r"%maxver)
	if isinstance(minver, str):
		minver = minver.split('.')
	if isinstance(maxver, str):
		maxver = maxver.split('.')
	if minver:
		minver = tuple(map(int, minver))
	if maxver:
		maxver = tuple(map(int, maxver))
	
	# Trick to be compatible python 2.x and 3.x
	try:
		u = unicode
	except NameError:
		u = str

	cmd = [self.env.CYTHON[0], '--version']
	o = u(check_output(cmd, stderr=STDOUT), 'UTF-8').strip()
	m = cython_ver_re.match(o)
	self.start_msg('Checking for cython version')
	if not m:
		self.end_msg('Not found', 'YELLOW')
		self.fatal("No version found")
	else:
		v = m.group(1)
		ver = tuple(map(int, v.split('.')))
		check = (not minver or minver <= ver) and (not maxver or maxver >= ver)
		self.to_log('  cython %s\n  -> %r\n' % (" ".join(log_s), v))
		if check:
			self.end_msg(v)
			self.env.CYTHON_VERSION = ver
		else:
			self.end_msg('wrong version %s' % v, 'YELLOW')
			self.fatal("Bad cython version (%s)"%v)
	return True


def configure(conf):
	if not conf.env.CYTHON:
		conf.fatal("cython_extra requires the tool 'cython' to be loaded first.")
	Configure.conf(check_cython_version)
	



#!/usr/bin/env python
from subprocess import check_output, call
import tarfile
import os

def get_template(ball, path):
    template = {}

    tfile = tarfile.open(ball, "r")

    reference = tfile.getmember(path)
    for attr in ('uid', 'gid', 'uname', 'gname', 'mtime'):
        template[attr] = getattr(reference, attr)

    tfile.close()

    return template

def add_files(ball, prefix, template, files):
    tfile = tarfile.open(ball, "a")

    for name, content in files:
        path = os.path.join(os.path.dirname(ball), name)
        if os.path.exists(path):
            os.unlink(path)

        fd = open(path, "wb+")
        fd.write(content)
        fd.close()

        tinfo = tfile.gettarinfo(arcname=os.path.join(prefix, name), name=path)
        for key, value in template.items():
            setattr(tinfo, key, value)

        fd = open(path, "rb")
        tfile.addfile(tinfo, fileobj=fd)
        fd.close()

    tfile.close()

VERSION = check_output(["git", "describe"]).decode().strip()
print("Found %s version" % VERSION)

# TODO: derive paths from submodule configuration
TUTORIAL_DIR="doc/tutorial"
S4_DIR="src/lib/s4"

PREFIX="xmms2-%s" % VERSION
PREFIX_TUTORIAL="%s/%s" % (PREFIX, TUTORIAL_DIR)
PREFIX_S4="%s/%s" % (PREFIX, S4_DIR)

DIST_DIR="dist"
DIST_XMMS2="%s/xmms2-%s.tar" % (DIST_DIR, VERSION)
DIST_XMMS2_BZ2="%s/xmms2-%s.tar.bz2" % (DIST_DIR, VERSION)
DIST_TUTORIAL="%s/xmms2-tutorial-%s.tar" % (DIST_DIR, VERSION)
DIST_S4="%s/xmms2-s4-%s.tar" % (DIST_DIR, VERSION)

if not os.path.exists(DIST_DIR):
    os.mkdir(DIST_DIR)

if os.path.exists(DIST_XMMS2):
    os.unlink(DIST_XMMS2)

if os.path.exists(DIST_XMMS2_BZ2):
    os.unlink(DIST_XMMS2_BZ2)

if os.path.exists(DIST_TUTORIAL):
    os.unlink(DIST_TUTORIAL)

if os.path.exists(DIST_S4):
    os.unlink(DIST_S4)

print("Create %s" % DIST_XMMS2)
call("git archive --format=tar --prefix=%s/ HEAD > %s" % (PREFIX, DIST_XMMS2), shell=True)

call("git submodule init", shell=True)
call("git submodule update", shell=True)
print("Create %s" % DIST_TUTORIAL)
call("git --git-dir=%s/.git archive --format=tar --prefix=%s/ HEAD > %s" % (TUTORIAL_DIR, PREFIX_TUTORIAL, DIST_TUTORIAL), shell=True)
print("Create %s" % DIST_S4)
call("git --git-dir=%s/.git archive --format=tar --prefix=%s/ HEAD > %s" % (S4_DIR, PREFIX_S4, DIST_S4), shell=True)

print("Append %s to %s" % (DIST_TUTORIAL, DIST_XMMS2))
call("tar -Af %s %s" % (DIST_XMMS2, DIST_TUTORIAL), shell=True)
print("Append %s to %s" % (DIST_S4, DIST_XMMS2))
call("tar -Af %s %s" % (DIST_XMMS2, DIST_S4), shell=True)

print("Append ChangeLong and hashed to %s" % DIST_XMMS2)
add_files(DIST_XMMS2, PREFIX, get_template(DIST_XMMS2, os.path.join(PREFIX, "wscript")), [
        ("xmms2-%s.ChangeLog" % VERSION, check_output("utils/gen-changelog.py")),
        ("checksums", check_output("utils/gen-tree-hashes.py"))
])

print("Compress %s" % DIST_XMMS2)
call("xz %s" % DIST_XMMS2, shell=True)



# XMMS2 - X Music Multiplexer System
# Copyright (C) 2003-2023 XMMS2 Team
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# python -m unittest discover -s src/clients/nycli -p *.py -v

from collections import defaultdict
import subprocess
import re
import time
import unittest

class CommandResult(object):
    def __init__(self, stdout, stderr, success):
        self.stdout = stdout
        self.stderr = stderr
        self.success = success

    def __repr__(self):
        return "stdout: %s\nstderr: %s\ncode: %d" % (self.stdout, self.stderr, self.success)

def cmd(*args):
    proc = subprocess.Popen(("xmms2",) + tuple(str(x) for x in args), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    return CommandResult(stdout, stderr, proc.returncode == 0)

class Properties(object):
    def __init__(self, by_key, by_source):
        self.by_key = by_key
        self.by_source = by_source

    def __repr__(self):
        result = ""
        for s,kv in self.by_source.items():
            for k,v in kv.iteritems():
                result += "[%s] %s = %s\n" % (s, k, v)
        return result.strip()

    @staticmethod
    def parse(text):
        by_key = defaultdict(dict)
        by_source = defaultdict(dict)
        pattern = re.compile("\[([^]]+)\] ([^ ]+) = (.+)")
        for line in text.split("\n"):
            line = line.strip()
            if not line:
                continue
            source, key, value = pattern.match(line).groups()
            by_key[key][source] = value
            by_source[source][key] = value

        return Properties(by_key, by_source)

    @staticmethod
    def fetch(mid=None):
        if mid is not None:
            return Properties.parse(cmd("info", "#%d" % mid).stdout)
        return Properties.parse(cmd("info").stdout)

class PlaylistEntry(object):
    def __init__(self, mid, artist, title, duration):
        self.mid = mid
        self.artist = artist
        self.title = title
        self.duration = duration

        minutes, seconds = duration.split(":")
        self.seconds = int(minutes) * 60 + int(seconds)

    def __repr__(self):
        return "[%d] %s %s (%s)" % (self.mid, self.artist, self.title, self.duration)

class Playlist(object):
    def __init__(self, entries, current_position):
        self.entries = entries
        self.current_position = current_position

    def __repr__(self):
        result = ""
        for position, entry in enumerate(self.entries):
            result += str(position == self.current_position) + " " + str(entry) + "\n"
        return result.strip()

    @staticmethod
    def parse_columns(lines):
        pass

    @staticmethod
    def parse_classic(lines):
        entries = []
        start_position = 0
        end_position = 0
        current_position = 0
        pattern = re.compile("([^[]+)\[([^/]+)/([^]]+)\] (.+) - (.+) \(([^)]+)\)")
        for line in lines:
            if len(line) == 0 or line.startswith("Total"):
                continue
            status, position, mid, artist, title, duration = pattern.match(line).groups()
            if status == "->":
                current_position = int(position)
            if start_position == 0:
                start_position = position
            end_position = position
            entries.append(PlaylistEntry(int(mid), artist, title, duration))
        return Playlist(entries, current_position)

    @staticmethod
    def parse(text):
        lines = text.split("\n")
        if len(lines) > 0:
            if lines[0].startswith("--[Result]"):
                return Playlist.parse_columns(lines)
            return Playlist.parse_classic(lines)
        raise Exception("Could not parse playlist output")

    @staticmethod
    def fetch(playlist=None):
        if playlist is not None:
            return Playlist.parse(cmd("list", "-p", playlist).stdout)
        return Playlist.parse(cmd("list").stdout)

class Collections(object):
    def __init__(self, entries, active):
        self.entries = entries
        self.active = active

    @staticmethod
    def parse(text):
        active = None
        entries = []
        for line in text.split("\n"):
            if line.startswith("* "):
                active = line[2:]
            entries.append(line[2:])
        return Collections(entries, active)

    @staticmethod
    def fetch_playlists(playlist=None):
        if playlist is not None:
            return Collections.parse(cmd("playlist", "list", playlist).stdout)
        return Collections.parse(cmd("playlist", "list").stdout)

    @staticmethod
    def fetch_collections():
        return Collections.parse(cmd("collections", "list").stdout)

class KeyValue(object):
    @staticmethod
    def parse(text, func=None):
        kv = {}
        for line in text.split("\n"):
            line = line.strip()
            if not line:
                continue
            key, value = line.split("=")
            if func is not None:
                try:
                    kv[key.strip()] = func(value.strip())
                except:
                    print "Could not marshall '%s' = '%s'" % (key, value)
            else:
                kv[key.strip()] = value.strip()
        return kv

class Volume(object):
    @staticmethod
    def fetch():
        return KeyValue.parse(cmd("server", "volume").stdout, int)

class Config(object):
    @staticmethod
    def fetch():
        return KeyValue.parse(cmd("server", "config").stdout)

class BasicPlaylist(unittest.TestCase):
    TEST_PLAYLIST_1 = "cli-test-1"
    TEST_PLAYLIST_2 = "cli-test-2"
    TEST_PLAYLIST_3 = "cli-test-3"
    TEST_PLAYLIST_4 = "cli-test-4"

    def setUp(self):
        cmd("playlist", "create", BasicPlaylist.TEST_PLAYLIST_1)
        cmd("playlist", "clear", BasicPlaylist.TEST_PLAYLIST_1)
        cmd("add", "-p", BasicPlaylist.TEST_PLAYLIST_1, "album:\"Modern Day City Symphony\"", "AND", "url~Music")

        cmd("playlist", "create", BasicPlaylist.TEST_PLAYLIST_2)
        cmd("playlist", "clear", BasicPlaylist.TEST_PLAYLIST_2)
        cmd("add", "-p", BasicPlaylist.TEST_PLAYLIST_2, "album:\"Modern Day City Symphony\"", "AND", "url~Music")

        self.previous_playlist = Collections.fetch_playlists().active

        cmd("playlist", "switch", BasicPlaylist.TEST_PLAYLIST_1)

    def tearDown(self):
        cmd("playlist", "switch", self.previous_playlist)

        playlists = (
            BasicPlaylist.TEST_PLAYLIST_1,
            BasicPlaylist.TEST_PLAYLIST_2,
            BasicPlaylist.TEST_PLAYLIST_3,
            BasicPlaylist.TEST_PLAYLIST_4
        )
        for playlist in playlists:
            cmd("playlist", "remove", playlist)


class TestPlaylist(BasicPlaylist):
    def testListEntries(self):
        # list active playlist
        playlist = Playlist.parse(cmd("list").stdout)
        self.assertEqual(15, len(playlist.entries))

        # list specific playlist
        playlist = Playlist.parse(cmd("list", "-p", TestPlaylist.TEST_PLAYLIST_2).stdout)
        self.assertEqual(15, len(playlist.entries))

        # list pattern
        playlist = Playlist.parse(cmd("list", "-p", TestPlaylist.TEST_PLAYLIST_2, "the").stdout)
        self.assertEqual(4, len(playlist.entries))

        # list positions around current position
        cmd("jump", "5")
        playlist = Playlist.parse(cmd("list", "2_2").stdout)
        self.assertEqual(5, playlist.current_position)
        self.assertEqual(5, len(playlist.entries))

        # list a number of positions
        playlist = Playlist.parse(cmd("list", "2,4,6,8").stdout)
        self.assertEqual(4, len(playlist.entries))

    def testClear(self):
        # clear active playlist
        self.assertEqual(15, len(Playlist.fetch(TestPlaylist.TEST_PLAYLIST_1).entries))
        cmd("playlist", "clear")
        self.assertEqual(0, len(Playlist.fetch(TestPlaylist.TEST_PLAYLIST_1).entries))

        # clear specified playlist
        self.assertEqual(15, len(Playlist.fetch(TestPlaylist.TEST_PLAYLIST_2).entries))
        cmd("playlist", "clear", TestPlaylist.TEST_PLAYLIST_2)
        self.assertEqual(0, len(Playlist.fetch(TestPlaylist.TEST_PLAYLIST_2).entries))

    def testRename(self):
        # rename active playlist
        cmd("playlist", "rename", TestPlaylist.TEST_PLAYLIST_3)
        self.assertTrue(TestPlaylist.TEST_PLAYLIST_3 in Collections.fetch_playlists().entries)

        # rename specified playlist
        cmd("playlist", "rename", "-p", TestPlaylist.TEST_PLAYLIST_2, TestPlaylist.TEST_PLAYLIST_4)
        self.assertTrue(TestPlaylist.TEST_PLAYLIST_4 in Collections.fetch_playlists().entries)

    def testCreate(self):
        # create with collision
        self.assertTrue("Error" in cmd("playlist", "create", TestPlaylist.TEST_PLAYLIST_1).stdout)

        # create new empty playlist
        self.assertTrue(cmd("playlist", "create", TestPlaylist.TEST_PLAYLIST_3).stdout.strip() == "")
        self.assertTrue("cli-test-3" in Collections.fetch_playlists().entries)
        self.assertEqual(0, len(Playlist.fetch(TestPlaylist.TEST_PLAYLIST_3).entries))

        # create new playlist and copy contents from another playlist
        self.assertTrue(cmd("playlist", "create", "-p", TestPlaylist.TEST_PLAYLIST_1, TestPlaylist.TEST_PLAYLIST_4).stdout.strip() == "")
        self.assertTrue("cli-test-3" in Collections.fetch_playlists().entries)
        self.assertEqual(15, len(Playlist.fetch(TestPlaylist.TEST_PLAYLIST_4).entries))

    def testShuffle(self):
        # shuffle active playlist
        before = list((x.artist, x.title) for x in Playlist.fetch().entries)
        cmd("playlist", "shuffle")
        after = list((x.artist, x.title) for x in Playlist.fetch().entries)
        self.assertNotEqual(before, after)
        self.assertEqual(set(before), set(after))

        # shuffle specific playlist
        before = list((x.artist, x.title) for x in Playlist.fetch(TestPlaylist.TEST_PLAYLIST_2).entries)
        cmd("playlist", "shuffle", TestPlaylist.TEST_PLAYLIST_2)
        after = list((x.artist, x.title) for x in Playlist.fetch(TestPlaylist.TEST_PLAYLIST_2).entries)
        self.assertNotEqual(before, after)
        self.assertEqual(set(before), set(after))

    def testSort(self):
        # sort active playlist by duration
        before = list(x.seconds for x in Playlist.fetch().entries)
        cmd("playlist", "sort", "duration")
        after = list(x.seconds for x in Playlist.fetch().entries)
        self.assertNotEqual(before, after)
        before.sort()
        self.assertEqual(before, after)

        # sort specific playlist by duration
        before = list(x.seconds for x in Playlist.fetch(TestPlaylist.TEST_PLAYLIST_2).entries)
        cmd("playlist", "sort", "-p", TestPlaylist.TEST_PLAYLIST_2, "duration")
        after = list(x.seconds for x in Playlist.fetch(TestPlaylist.TEST_PLAYLIST_2).entries)
        self.assertNotEqual(before, after)
        before.sort()
        self.assertEqual(before, after)

    def testClear(self):
        before = Playlist.fetch().entries
        cmd("playlist", "clear")
        after = Playlist.fetch().entries
        self.assertNotEqual(len(before), len(after))
        self.assertEqual(0, len(after))

        before = Playlist.fetch(TestPlaylist.TEST_PLAYLIST_2).entries
        cmd("playlist", "clear", TestPlaylist.TEST_PLAYLIST_2)
        after = Playlist.fetch(TestPlaylist.TEST_PLAYLIST_2).entries
        self.assertNotEqual(len(before), len(after))
        self.assertEqual(0, len(after))

    def testSwitch(self):
        # switch to second test playlist
        self.assertEqual(TestPlaylist.TEST_PLAYLIST_1, Collections.fetch_playlists().active)
        cmd("playlist", "switch", TestPlaylist.TEST_PLAYLIST_2)
        self.assertEqual(TestPlaylist.TEST_PLAYLIST_2, Collections.fetch_playlists().active)
        # and switch back again
        cmd("playlist", "switch", TestPlaylist.TEST_PLAYLIST_1)
        self.assertEqual(TestPlaylist.TEST_PLAYLIST_1, Collections.fetch_playlists().active)

    def testList(self):
        # list playlists
        result = cmd("playlist", "list")
        self.assertTrue(TestPlaylist.TEST_PLAYLIST_1 in result.stdout)
        self.assertTrue(TestPlaylist.TEST_PLAYLIST_2 in result.stdout)

        # list playlists + hidden playlists
        result = cmd("playlist", "list", "-a")
        self.assertTrue(TestPlaylist.TEST_PLAYLIST_1 in result.stdout)
        self.assertTrue(TestPlaylist.TEST_PLAYLIST_2 in result.stdout)
        self.assertTrue("_active" in result.stdout)

class TestServer(BasicPlaylist):
    def testBrowse(self):
        self.assertTrue("file://" in cmd("server", "browse", "file:///").stdout)

    def testPlugins(self):
        self.assertTrue("segment" in cmd("server", "plugins").stdout)

    def testVolume(self):
        original = Volume.fetch()

        cmd("server", "volume", "100")
        volume = Volume.fetch()
        self.assertEqual(set([100]), set(volume.values()))

        cmd("server", "volume", "0")
        volume = Volume.fetch()
        self.assertEqual(set([0]), set(volume.values()))

        cmd("server", "volume", "+10")
        volume = Volume.fetch()
        self.assertEqual(set([10]), set(volume.values()))

        cmd("server", "volume", "-5")
        volume = Volume.fetch()
        self.assertEqual(set([5]), set(volume.values()))

        volume = Volume.fetch()
        channel = volume.iterkeys().next()
        cmd("server", "volume", "-c", channel, "42")
        volume = Volume.fetch()
        self.assertTrue(42 in set(volume.values()))

        for channel, volume in original.iteritems():
            cmd("server", "volume", "-c", channel, volume)

    def testSync(self):
        self.assertTrue("Error" not in cmd("server", "sync").stdout)

    def testStats(self):
        result = cmd("server", "stats")
        self.assertTrue("uptime" in result.stdout)
        self.assertTrue("version" in result.stdout)

    def testRehash(self):
        entry = iter(Playlist.fetch().entries).next()

        before = Properties.fetch(entry.mid)
        for source in before.by_key["artist"].iterkeys():
            cmd("server", "property", "-S", source, "-D", entry.mid, "artist")

        after = Properties.fetch(entry.mid)
        self.assertNotEqual(before.by_key["artist"], after.by_key["artist"])
        cmd("server", "rehash", "#%d" % entry.mid)

        # TODO: replace with await(lambda) as rehash is async
        time.sleep(0.5)

        after = Properties.fetch(entry.mid)
        self.assertEqual(before.by_key["artist"], after.by_key["artist"])

    def testConfig(self):
        before = Config.fetch()
        self.assertTrue("output.buffersize" in before.keys())

        # change config setting
        cmd("server", "config", "output.buffersize", 512)
        after = Config.fetch()
        self.assertNotEqual(before["output.buffersize"], after["output.buffersize"])

        # restore config setting
        cmd("server", "config", "output.buffersize", before["output.buffersize"])
        after = Config.fetch()
        self.assertEqual(before["output.buffersize"], after["output.buffersize"])

        # print single
        result = cmd("server", "config", "output.buffersize")
        self.assertTrue("output.buffersize" in result.stdout)

        # print match
        result = cmd("server", "config", "*buffer*")
        self.assertTrue("output.buffersize" in result.stdout)


class TestPlayback(BasicPlaylist):
    def testPlaybackToggle(self):
        self.assertTrue(cmd("stop").success)
        self.assertEquals(cmd("current", "-f", "${playback_status}").stdout.strip(), "Stopped")
        self.assertTrue(cmd("play").success)
        self.assertEquals(cmd("current", "-f", "${playback_status}").stdout.strip(), "Playing")
        self.assertTrue(cmd("stop").success)

    def testPlaybackSeek(self):
        def pos():
            minutes, seconds = cmd("current", "-f", "${playtime}").stdout.strip().split(":")
            return int(minutes) * 60 + int(seconds)

        self.assertTrue(cmd("stop").success)
        self.assertTrue(cmd("play").success)
        # TODO: Use await() here
        time.sleep(0.2)
        self.assertEquals(0, pos())
        self.assertTrue(cmd("seek", "20").success)
        self.assertEquals(20, pos())
        self.assertTrue(cmd("seek", "0").success)
        # TODO: Use await() here
        time.sleep(0.2)
        self.assertTrue(pos() < 20)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import sys
import genipc
from indenter import Indenter

# Python 2.x compat
try:
	basestring
except:
	basestring = str

# dictionary mapping of types in the xml to C type aliases
c_type_map = {
	'string': 'const char *',
	'int': 'gint32',
	'enum-value': lambda ename: 'gint32',
	'collection':'xmmsv_t *',
	'binary':'GString *',
	'list':'xmmsv_t *',
	'dictionary':'xmmsv_t *',
	'unknown' : 'xmmsv_t *',
}
def get_type(typeinfo):
	if isinstance(typeinfo, basestring):
		return c_type_map[typeinfo]
	return c_type_map[typeinfo[0]](*(typeinfo[1:]))

c_getter_map = {
	'int': 'xmmsv_get_int',
	'string': 'xmmsv_get_string',
	'enum-value': 'xmmsv_get_int',
	'list': None,
	'dictionary': None,
	'collection': None,
	'binary': 'xmms_bin_to_gstring',
	'unknown' : None,
}
def get_getter(typeinfo):
	if not isinstance(typeinfo, basestring):
		typeinfo = typeinfo[0]
	return c_getter_map[typeinfo]

c_creator_map = {
	'int': 'xmmsv_new_int',
	'string': 'xmms_convert_and_kill_string',
	'enum-value': 'xmmsv_new_int',
	'list': None,
	'dictionary': None,
	'collection': None,
	'binary': None,
	'unknown' : None,
}
def get_creator(typeinfo):
	if not isinstance(typeinfo, basestring):
		typeinfo = typeinfo[0]
	return c_creator_map[typeinfo]

c_nullable_type_map = {
	'int': False,
	'string': "gchar *",
	'enum-value': False,
	'list': "xmmsv_t *",
	'dictionary': "xmmsv_t *",
	'collection': "xmmsv_t *",
	'binary': "GString *",
	'unknown' : "xmmsv_t *",
}
def get_nullable(typeinfo):
	if not isinstance(typeinfo, basestring):
		typeinfo = typeinfo[0]
	return c_nullable_type_map[typeinfo]

c_xmmsv_type_t_map = {
	'int': 'XMMSV_TYPE_INT64',
	'string': 'XMMSV_TYPE_STRING',
	'enum-value': 'XMMSV_TYPE_INT64',
	'list': 'XMMSV_TYPE_LIST',
	'dictionary': 'XMMSV_TYPE_DICT',
	'collection': 'XMMSV_TYPE_COLL',
	'binary': 'XMMSV_TYPE_BINDATA'
}
def get_xmmsv_type(typeinfo):
	if not isinstance(typeinfo, basestring):
		typeinfo = typeinfo[0]
	return c_xmmsv_type_t_map[typeinfo]
def has_xmmsv_type(typeinfo):
	if not isinstance(typeinfo, basestring):
		typeinfo = typeinfo[0]
	return typeinfo in c_xmmsv_type_t_map

def _enum_value(member):
	return 'XMMS_%s' % member.fullname()

def build(object_name, c_type):
	ipc = genipc.parse_xml('../src/ipc.xml')

	Indenter.printline('/* This code is automatically generated by genipc_server.py. Do not edit. */')
	Indenter.printline()
	Indenter.printline("#include <xmmsc/xmmsv.h>")

	for object in ipc.objects:
		if object.name == object_name:
			for method in object.methods:
				emit_method_define_code(object, method, c_type)

			Indenter.printline()
			Indenter.printline('static void')
			Indenter.printline('xmms_%s_register_ipc_commands (xmms_object_t *%s_object)' % (object.name, object.name))
			Indenter.enter('{')

			Indenter.printline('xmms_ipc_object_register (%s, %s_object);' % (_enum_value(object.id), object.name))
			Indenter.printline()

			for method in object.methods:
				emit_method_add_code(object, method)

			Indenter.printline()

			for broadcast in object.broadcasts:
				Indenter.printline('xmms_ipc_broadcast_register (%s_object, %s);' % (object.name, _enum_value(broadcast.id)))

			Indenter.printline()

			for signal in object.signals:
				Indenter.printline('xmms_ipc_signal_register (%s_object, %s);' % (object.name, _enum_value(signal.id)))

			Indenter.leave('}')

			Indenter.printline()
			Indenter.printline('static void')
			Indenter.printline('xmms_%s_unregister_ipc_commands (void)' % object.name)
			Indenter.enter('{')

			for broadcast in object.broadcasts:
				Indenter.printline('xmms_ipc_broadcast_unregister (%s);' % _enum_value(broadcast.id))

			Indenter.printline()

			for signal in object.signals:
				Indenter.printline('xmms_ipc_signal_unregister (%s);' % _enum_value(signal.id))

			Indenter.printline()
			Indenter.printline('xmms_ipc_object_unregister (%s);' % _enum_value(object.id))
			Indenter.leave('}')

def method_name_to_cname(n):
	return "__int_xmms_cmd_%s" % n

def emit_method_define_code(object, method, c_type):
	full_method_name = 'xmms_%s_client_%s' % (object.name, method.name)

	Indenter.printline("static void")
	Indenter.printline("%s (xmms_object_t *object, xmms_object_cmd_arg_t *arg)" % method_name_to_cname (method.name))
	Indenter.enter("{")

	if method.arguments:
		Indenter.printline("xmmsv_t *t;")

	Indenter.enter("if (xmmsv_list_get_size (arg->args) != %d) {" % len(method.arguments))
	Indenter.printline('XMMS_DBG ("Wrong number of arguments to %s (%%d)", xmmsv_list_get_size (arg->args));' % method.name)
	Indenter.printline('xmms_error_set (&arg->error, XMMS_ERROR_INVAL, "Wrong number of arguments to %s");' % method.name)
	Indenter.printline('return;')
	Indenter.leave("}")

	for i, a in enumerate(method.arguments):
		Indenter.printline("%s argval%d;" % (get_type(a.type[0]), i))

	Indenter.printline()

	for i, a in enumerate(method.arguments):
		Indenter.enter("if (!xmmsv_list_get (arg->args, %d, &t)) {" % i)
		Indenter.printline('XMMS_DBG ("Missing arg %d in %s");' % (i, method.name))
		Indenter.printline('xmms_error_set (&arg->error, XMMS_ERROR_INVAL, "Missing arg %d in %s");' % (i, method.name))
		Indenter.printline('return;')
		Indenter.leave("}")

		if a.type[0] == 'list' and len(a.type) > 1 and has_xmmsv_type(a.type[1]):
			Indenter.enter('if (!xmmsv_list_restrict_type (t, %s)) {' % (get_xmmsv_type(a.type[1])))
			Indenter.printline('XMMS_DBG("Wrong list content (not %s) for arg %d in %s.");' % (a.type[1], i, method.name))
			Indenter.printline('xmms_error_set (&arg->error, XMMS_ERROR_INVAL, "Wrong list content (not %s) for arg %d in %s.");' % (a.type[1], i, method.name))
			Indenter.leave('}')

		if get_getter(a.type[0]) is None:
			Indenter.printline("argval%d = t;" % i)
		else:
			Indenter.enter("if (!%s (t, &argval%d)) {" % (get_getter(a.type[0]), i))
			Indenter.printline('XMMS_DBG ("Error parsing arg %d in %s");' % (i, method.name))
			Indenter.printline('xmms_error_set (&arg->error, XMMS_ERROR_INVAL, "Error parsing arg %d in %s");' % (i, method.name))
			Indenter.printline("return;")
			Indenter.leave("}")


	Indenter.printline()

	# Assemble a list of arguments for the function call
	args = []

	args.append("(%s) object" % c_type)
	args.extend("argval%d" % i for i in range(len(method.arguments)))

	if method.need_client:
		args.append("arg->client")
	if method.need_cookie:
		args.append("arg->cookie")

	args.append("&arg->error")

	funccall = "%s (%s)" % (full_method_name, ", ".join(args))
	if method.return_value:
		if get_creator(method.return_value.type[0]) is None:
			Indenter.printline("arg->retval = %s;" % (funccall))
		else:
			if get_nullable(method.return_value.type[0]):
				Indenter.printline("%s retval = %s;" % (get_nullable(method.return_value.type[0]), funccall))
				Indenter.enter("if (retval != NULL) {")
				Indenter.printline("arg->retval = %s (retval);" % get_creator(method.return_value.type[0]))
				Indenter.leave("}")
			else:
				Indenter.printline("arg->retval = %s (%s);" % (get_creator(method.return_value.type[0]), funccall))
	else:
		Indenter.printline("%s;" % funccall)
		if method.noreply:
			Indenter.printline("arg->retval = NULL;")
		else:
			Indenter.printline("arg->retval = xmmsv_new_none ();")


	Indenter.leave("}")

	Indenter.printline()
	Indenter.printline()


def emit_method_add_code(object, method):
	Indenter.printline('xmms_object_cmd_add (%s_object, %s, %s);' % (object.name, _enum_value(method.id), method_name_to_cname (method.name)))



"""
Python bindings for xmmsv manipulations.
"""

from xmmsutils cimport *
from cpython.bytes cimport PyBytes_FromStringAndSize
cimport cython
from cxmmsvalue cimport *

VALUE_TYPE_NONE   = XMMSV_TYPE_NONE
VALUE_TYPE_ERROR  = XMMSV_TYPE_ERROR
VALUE_TYPE_INT64  = XMMSV_TYPE_INT64
VALUE_TYPE_FLOAT  = XMMSV_TYPE_FLOAT
VALUE_TYPE_STRING = XMMSV_TYPE_STRING
VALUE_TYPE_COLL   = XMMSV_TYPE_COLL
VALUE_TYPE_BIN    = XMMSV_TYPE_BIN
VALUE_TYPE_LIST   = XMMSV_TYPE_LIST
VALUE_TYPE_DICT   = XMMSV_TYPE_DICT

COLLECTION_TYPE_REFERENCE    = XMMS_COLLECTION_TYPE_REFERENCE
COLLECTION_TYPE_UNIVERSE     = XMMS_COLLECTION_TYPE_UNIVERSE
COLLECTION_TYPE_UNION        = XMMS_COLLECTION_TYPE_UNION
COLLECTION_TYPE_INTERSECTION = XMMS_COLLECTION_TYPE_INTERSECTION
COLLECTION_TYPE_COMPLEMENT   = XMMS_COLLECTION_TYPE_COMPLEMENT
COLLECTION_TYPE_HAS          = XMMS_COLLECTION_TYPE_HAS
COLLECTION_TYPE_MATCH        = XMMS_COLLECTION_TYPE_MATCH
COLLECTION_TYPE_TOKEN        = XMMS_COLLECTION_TYPE_TOKEN
COLLECTION_TYPE_EQUALS       = XMMS_COLLECTION_TYPE_EQUALS
COLLECTION_TYPE_NOTEQUAL     = XMMS_COLLECTION_TYPE_NOTEQUAL
COLLECTION_TYPE_SMALLER      = XMMS_COLLECTION_TYPE_SMALLER
COLLECTION_TYPE_SMALLEREQ    = XMMS_COLLECTION_TYPE_SMALLEREQ
COLLECTION_TYPE_GREATER      = XMMS_COLLECTION_TYPE_GREATER
COLLECTION_TYPE_GREATEREQ    = XMMS_COLLECTION_TYPE_GREATEREQ
COLLECTION_TYPE_ORDER        = XMMS_COLLECTION_TYPE_ORDER
COLLECTION_TYPE_LIMIT        = XMMS_COLLECTION_TYPE_LIMIT
COLLECTION_TYPE_MEDIASET     = XMMS_COLLECTION_TYPE_MEDIASET
COLLECTION_TYPE_IDLIST       = XMMS_COLLECTION_TYPE_IDLIST


from propdict import PropDict # xmmsclient.propdict


cdef xmmsv_t *create_native_value(value) except NULL:
	cdef xmmsv_t *ret = NULL
	cdef xmmsv_t *v
	cdef XmmsValue xv
	cdef Collection c

	if value is None:
		ret = xmmsv_new_none()
	elif isinstance(value, XmmsValue):
		xv = value
		ret = xmmsv_copy(xv.val) # Prevent side effects.
	elif isinstance(value, Collection):
		c = value
		ret = xmmsv_copy(<xmmsv_t *>c.coll) #Prevent side effects.
	elif isinstance(value, int):
		ret = xmmsv_new_int(value)
	elif isinstance(value, (unicode, str)):
		s = from_unicode(value)
		ret = xmmsv_new_string(s)
	elif hasattr(value, "keys"):
		ret = xmmsv_new_dict()
		for key in value.keys():
			k = from_unicode(key)
			v = create_native_value(value[key])
			xmmsv_dict_set(ret, k, v)
			xmmsv_unref(v)
	else:
		try:
			it = iter(value)
		except TypeError:
			raise TypeError("Type '%s' has no corresponding native value type"
					% value.__class__.__name__)
		else:
			ret = xmmsv_new_list()
			for item in value:
				v = create_native_value(item)
				xmmsv_list_append(ret, v)
				xmmsv_unref(v) # the list took its own reference
	return ret

class XmmsError(Exception): pass

cdef class XmmsValue:
	#cdef object sourcepref
	#cdef xmmsv_t *val
	#cdef int ispropdict

	def __cinit__(self):
		# Prevent stupid segfault from user missuse.
		self.val = xmmsv_new_none()
		self.ispropdict = 0

	def __init__(self, sourcepref = None, pyval = None):
		cdef xmmsv_t *v
		self.sourcepref = sourcepref

		if pyval is not None: # None value already created.
			XmmsValue.set_pyval(self, pyval) # Use cdef version of the function

	def __dealloc__(self):
		if self.val != NULL:
			xmmsv_unref(self.val)
			self.val = NULL

	cdef set_value(self, xmmsv_t *value, int ispropdict = -1):
		if self.val != NULL:
			xmmsv_unref(self.val)
		xmmsv_ref(value)
		self.val = value
		if ispropdict >= 0:
			self.ispropdict = ispropdict

	cpdef set_pyval(self, pyval):
		cdef xmmsv_t *v
		v = create_native_value(pyval)
		self.set_value(v)
		xmmsv_unref(v)

	cpdef get_type(self):
		"""
		Return the type of the value.

		:return: The return value is one of the VALUE_TYPE_* constants.
		"""
		return xmmsv_get_type(self.val)

	cpdef iserror(self):
		"""
		:deprecated: Use `XmmsValue.is_error` instead.
		:return: Whether the value represents an error or not.
		:rtype: bool
		"""
		return self.is_error()

	cpdef is_error(self):
		"""
		:return: Whether the value represents an error or not.
		:rtype: Boolean
		"""
		return xmmsv_is_type(self.val, XMMSV_TYPE_ERROR)

	cpdef get_error(self):
		"""
		:return: Error string from the result.
		:rtype: String
		"""
		cdef char *ret = NULL
		if not xmmsv_get_error(self.val, <const_char **>&ret):
			raise ValueError("Failed to retrieve value")
		return XmmsError(to_unicode(ret))

	cpdef get_int(self):
		"""
		Get data from the result structure as an int.

		:return: An integer from the result.
		:rtype: int
		"""
		cdef int64_t ret = 0
		if not xmmsv_get_int64(self.val, &ret):
			raise ValueError("Failed to retrieve value")
		return ret

	cpdef get_float(self):
		"""
		Get data from the result structure as an float.

		:return: A float from the result.
		:rtype: float
		"""
		cdef float ret = 0
		if not xmmsv_get_float(self.val, &ret):
			raise ValueError("Failed to retrieve value")
		return ret

	cpdef get_string(self):
		"""
		Get data from the result structure as a string.

		:return: A string from the result.
		:rtype: string
		"""
		cdef char *ret = NULL
		if not xmmsv_get_string(self.val, <const_char **>&ret):
			raise ValueError("Failed to retrieve value")
		return to_unicode(ret)

	cpdef get_bin(self):
		"""
		Get data from the result structure as binary data.

		:return: The binary data as bytes.
		:rtype: bytes
		"""
		cdef char *ret = NULL
		cdef unsigned int rlen = 0
		if not xmmsv_get_bin(self.val, <const_uchar **>&ret, &rlen):
			raise ValueError("Failed to retrieve value")
		return PyBytes_FromStringAndSize(<char *>ret, rlen)

	cpdef get_coll(self):
		"""
		Get data from the result structure as a Collection.

		:rtype: Collection
		"""
		if not xmmsv_is_type(self.val, XMMSV_TYPE_COLL):
			raise ValueError("The value is not a collection")
		return create_coll(self.val)

	cpdef get_dict(self):
		"""
		:return: A dictionary containing media info.
		"""
		return dict([(k, v.value()) for k,v in self.get_dict_iter()])

	cpdef get_dict_iter(self):
		"""
		:return: An iterator on dict items ((key, value) pairs).
		"""
		return XmmsDictIter(self)

	cpdef get_propdict(self):
		"""
		:return: A source dict.
		"""
		ret = PropDict(self.sourcepref)
		for key, values in self.get_dict_iter():
			for source, value in values.get_dict_iter():
				ret[(source, key)] = value.value()
		return ret

	cpdef get_list(self):
		"""
		:return: A list of dicts from the result structure.
		"""
		return [v.value() for v in self.get_list_iter()]

	cpdef get_list_iter(self):
		"""
		:return: An iterator on a list.
		"""
		return XmmsListIter(self)

	cpdef value(self):
		"""
		This can be used instead of most get_* functions in this class.

		:return: Return value of appropriate data type contained in this result.
		"""
		cdef xmmsv_type_t vtype
		vtype = self.get_type()

		if vtype == XMMSV_TYPE_NONE:
			return None
		elif vtype == XMMSV_TYPE_ERROR:
			return self.get_error()
		elif vtype == XMMSV_TYPE_INT64:
			return self.get_int()
		elif vtype == XMMSV_TYPE_FLOAT:
			return self.get_float()
		elif vtype == XMMSV_TYPE_STRING:
			return self.get_string()
		elif vtype == XMMSV_TYPE_COLL:
			return self.get_coll()
		elif vtype == XMMSV_TYPE_BIN:
			return self.get_bin()
		elif vtype == XMMSV_TYPE_LIST:
			return self.get_list()
		elif vtype == XMMSV_TYPE_DICT:
			if self.ispropdict:
				return self.get_propdict()
			return self.get_dict()
		else:
			raise TypeError("Unknown value type from the server: %d" % vtype)

	cpdef copy(self, cls=None):
		if cls is None:
			cls = XmmsValue
		elif not issubclass(cls, XmmsValue):
			raise TypeError("Custom class must be a subclass of XmmsValue")
		return cls(sourcepref = self.sourcepref, pyval = self)

	def __iter__(self):
		vtype = self.get_type()

		if vtype == XMMSV_TYPE_LIST:
			return self.get_list_iter()
		elif vtype == XMMSV_TYPE_DICT:
			return self.get_dict_iter()
		elif vtype == XMMSV_TYPE_STRING:
			return iter(self.get_string())
		elif vtype == XMMSV_TYPE_BIN:
			return iter(self.get_bin())
		raise TypeError("Value type not iterable")



cdef class XmmsValueC2C(XmmsValue):
	property sender:
		def __get__(self):
			if xmmsv_get_type(self.val) != XMMSV_TYPE_DICT:
				return -1
			return xmmsv_c2c_message_get_sender(self.val)

	property id:
		def __get__(self):
			if xmmsv_get_type(self.val) != XMMSV_TYPE_DICT:
				return -1
			return xmmsv_c2c_message_get_id(self.val)

	property destination:
		def __get__(self):
			if xmmsv_get_type(self.val) != XMMSV_TYPE_DICT:
				return -1
			return xmmsv_c2c_message_get_destination(self.val)

	property payload:
		def __get__(self):
			cdef xmmsv_t *payload
			cdef XmmsValue value
			if xmmsv_get_type(self.val) != XMMSV_TYPE_DICT:
				return None
			payload = xmmsv_c2c_message_get_payload(self.val)
			if payload == NULL:
				return None
			value = XmmsValue()
			value.set_value(payload)
			return value



cdef class XmmsListIter:
	#cdef object sourcepref
	#cdef xmmsv_t *val
	#cdef xmmsv_list_iter_t *it

	def __cinit__(self):
		self.val = NULL
		self.it = NULL

	def __dealloc__(self):
		if self.it != NULL:
			xmmsv_list_iter_explicit_destroy(self.it)
		if self.val != NULL:
			xmmsv_unref(self.val)

	def __init__(self, XmmsValue value):
		if value.get_type() != XMMSV_TYPE_LIST:
			raise TypeError("The value is not a list.")
		self.val = xmmsv_ref(value.val)
		if not xmmsv_get_list_iter(self.val, &self.it):
			raise RuntimeError("Failed to initialize the iterator.")
		self.sourcepref = value.sourcepref

	def __iter__(self):
		return self

	def __next__(self):
		cdef xmmsv_t *val = NULL
		cdef XmmsValue v

		if not xmmsv_list_iter_valid(self.it):
			raise StopIteration()

		if not xmmsv_list_iter_entry(self.it, &val):
			raise RuntimeError("Failed to retrieve list entry.")
		v = XmmsValue(self.sourcepref)
		v.set_value(val)

		xmmsv_list_iter_next(self.it)
		return v


cdef class XmmsDictIter:
	#cdef object sourcepref
	#cdef xmmsv_t *val
	#cdef xmmsv_dict_iter_t *it

	def __cinit__(self):
		self.val = NULL
		self.it = NULL

	def __dealloc__(self):
		if self.it != NULL:
			xmmsv_dict_iter_explicit_destroy(self.it)
		if self.val != NULL:
			xmmsv_unref(self.val)

	def __init__(self, XmmsValue value):
		if value.get_type() != XMMSV_TYPE_DICT:
			raise TypeError("The value is not a dict.")
		self.val = xmmsv_ref(value.val)
		if not xmmsv_get_dict_iter(self.val, &self.it):
			raise RuntimeError("Failed to initialize the iterator.")
		self.sourcepref = value.sourcepref

	def __iter__(self):
		return self

	def __next__(self):
		cdef char *key = NULL
		cdef xmmsv_t *val = NULL
		cdef XmmsValue v

		if not xmmsv_dict_iter_valid(self.it):
			raise StopIteration()

		if not xmmsv_dict_iter_pair(self.it, <const_char **>&key, &val):
			raise RuntimeError("Failed to retrieve dict item")
		v = XmmsValue(self.sourcepref)
		v.set_value(val)

		xmmsv_dict_iter_next(self.it)
		return (to_unicode(key), v)


cdef class CollectionRef:
	def __cinit__(self):
		self.coll = NULL

	def __dealloc__(self):
		if self.coll != NULL:
			xmmsv_unref(<xmmsv_t *>self.coll)
		self.coll = NULL

	cdef set_collection(self, xmmsv_t *coll):
		if self.coll != NULL:
			xmmsv_unref(<xmmsv_t *>self.coll)
		xmmsv_ref(<xmmsv_t *>coll)
		self.coll = coll


cdef _idlist_types = (
		XMMS_COLLECTION_TYPE_IDLIST,
		)
cdef class Collection(CollectionRef):
	#cdef object _attributes
	#cdef object _operands
	#cdef object _idlist

	cdef init_idlist(self):
		if self._idlist is None and self.coll != NULL and xmmsv_coll_get_type(self.coll) in _idlist_types:
			self._idlist = CollectionIDList(self)

	cdef init_attributes(self):
		if self._attributes is None and self.coll != NULL:
			self._attributes = CollectionAttributes(self)

	cdef init_operands(self):
		if self._operands is None and self.coll != NULL:
			self._operands = CollectionOperands(self)

	property attributes:
		def __get__(self):
			self.init_attributes()
			return self._attributes
		def __set__(self, attrs):
			self.init_attributes()
			if self._attributes is attrs:
				pass
			else:
				self._attributes.clear()
				self._attributes.update(attrs)
	property operands:
		def __get__(self):
			self.init_operands()
			return self._operands
		def __set__(self, ops):
			self.init_operands()
			if self._operands is ops:
				pass
			else:
				self._operands.clear()
				self._operands.extend(ops)
	property ids:
		def __get__(self):
			self.init_idlist()
			return self._idlist
		def __set__(self, ids):
			self.init_idlist()
			if self._idlist is None:
				raise TypeError("Can't set idlist for this type of collection")
			elif self._idlist is ids:
				pass
			else:
				self._idlist.clear()
				self._idlist.extend(ids)

	cpdef copy(self):
		cdef xmmsv_t *newcoll = xmmsv_copy(<xmmsv_t *>self.coll)
		coll = create_coll(newcoll)
		xmmsv_unref(newcoll)
		return coll

	def __repr__(self):
		atr = []
		operands = []
		if self.ids is not None:
			atr.append("ids=%r" % self.ids)
		for o in self.operands:
			operands.append(repr(o))
		if operands:
			atr.append("operands=[%s]" % ", ".join(operands))
		for k,v in self.attributes.iteritems():
			atr.append("%s=%s" % (k, repr(v)))
		return "%s(%s)" % (self.__class__.__name__, ", ".join(atr))

	def __or__(self, other): # |
		if not (isinstance(self, Collection) and isinstance(other, Collection)):
			raise TypeError("Unsupported operand type(s) for |: %r and %r"
				% (self.__class__.__name__, other.__class__.__name__))
		u = Union()
		for op in (self, other):
			if isinstance(op, Union):
				u.operands.extend(op.operands)
			else:
				u.operands.append(op)
		return u

	def __and__(self, other): # &
		if not (isinstance(self, Collection) and isinstance(other, Collection)):
			raise TypeError("Unsupported operand type(s) for &: %r and %r"
				% (self.__class__.__name__, other.__class__.__name__))
		i = Intersection()
		for op in (self, other):
			if isinstance(op, Intersection):
				i.operands.extend(op.operands)
			else:
				i.operands.append(op)
		return i

	def __invert__(self): #~
		if isinstance(self, Complement):
			return self.operands[0]
		return Complement(self)


cdef class AttributesIterator:
	#cdef xmmsv_dict_iter_t *diter
	#cdef int itertype
	#cdef CollectionAttributes attr

	def __cinit__(self):
		self.diter = NULL
		self.itertype = 0

	def __dealloc__(self):
		if self.diter != NULL:
			xmmsv_dict_iter_explicit_destroy(self.diter)

	def __init__(self, CollectionAttributes attributes, AttributesIterType itertype):
		cdef xmmsv_dict_iter_t *it = NULL
		if attributes is None:
			raise TypeError("__init__() argument must not be None")
		if not xmmsv_get_dict_iter(xmmsv_coll_attributes_get(attributes.coll), &it):
			raise RuntimeError("Failed to get fict iterator")
		xmmsv_dict_iter_first(it)
		self.diter = it
		self.itertype = itertype

	def __iter__(self):
		return self

	def __next__(self):
		cdef char *key = NULL
		cdef xmmsv_t *value = NULL
		if self.diter != NULL and xmmsv_dict_iter_valid(self.diter):
			xmmsv_dict_iter_pair(self.diter, <const_char **>&key, &value)
			k = to_unicode(key)
			if self.itertype == ITER_KEYS:
				ret = k
			else:
				v = XmmsValue()
				v.set_value(value)
				if self.itertype == ITER_VALUES:
					ret = v.values
				elif self.itertype == ITER_ITEMS:
					ret = (k, v.value())
				elif self.itertype == ITER_XVALUES:
					ret = v
				elif self.itertype == ITER_XITEMS:
					ret = (k, v)
				else:
					# Should never be reached
					raise RuntimeError("Unknown iter type")
			xmmsv_dict_iter_next(self.diter)
			return ret
		raise StopIteration()

	cpdef reset(self):
		xmmsv_dict_iter_first(self.diter)


cdef class CollectionAttributes(CollectionRef):
	def __init__(self, Collection c):
		if c.coll == NULL:
			raise RuntimeError("Uninitialized collection")
		self.set_collection(c.coll)

	cpdef get_dict(self):
		"""Get a copy of attributes dict"""
		return dict(self.iteritems())

	cpdef iterkeys(self):
		return AttributesIterator(self, ITER_KEYS)

	cpdef keys(self):
		return list(self.iterkeys())

	cpdef itervalues(self):
		return AttributesIterator(self, ITER_VALUES)

	cpdef iterxvalues(self):
		return AttributesIterator(self, ITER_XVALUES)

	cpdef values(self):
		return list(self.itervalues())

	cpdef xvalues(self):
		return list(self.iterxvalues())

	cpdef iteritems(self):
		return AttributesIterator(self, ITER_ITEMS)

	cpdef iterxitems(self):
		return AttributesIterator(self, ITER_XITEMS)

	cpdef items(self):
		return list(self.iteritems())

	cpdef xitems(self):
		return list(self.iterxitems())

	def __contains__(self, name):
		cdef xmmsv_t *value = NULL
		return xmmsv_coll_attribute_get_value(self.coll, name, &value)

	def __iter__(self):
		return self.iterkeys()

	def __repr__(self):
		return repr(self.get_dict())

	def __str__(self):
		return str(self.get_dict())

	def __getitem__(self, name):
		return self.xget(name).value()

	def __setitem__(self, name, value):
		cdef xmmsv_t *v
		n = from_unicode(name)
		v = create_native_value(value)
		xmmsv_coll_attribute_set_value(self.coll, <char *>n, v)
		xmmsv_unref(v)

	def __delitem__(self, name):
		n = from_unicode(name)
		if not xmmsv_coll_attribute_remove(self.coll, <char *>n):
			raise KeyError("The attribute '%s' doesn't exist" % name)

	def get(self, name, default = None):
		try:
			return self.xget(name).value()
		except KeyError:
			return default

	def xget(self, name):
		cdef xmmsv_t *value = NULL
		cdef XmmsValue xv
		nam = from_unicode(name)
		if not xmmsv_coll_attribute_get_value(self.coll, <char *>nam, &value):
			raise KeyError("The attribute '%s' doesn't exist" % name)
		xv = XmmsValue()
		xv.set_value(value)
		return xv

	cpdef clear(self):
		for k in self.keys():
			xmmsv_coll_attribute_remove(self.coll, k)

	def update(self, *a, **kargs):
		for d in a:
			if hasattr(d, 'keys') and hasattr(d.keys, '__call__'):
				for k in d.keys():
					self[k] = d[k]
			else:
				for k, v in d:
					self[k] = v
		for k in kargs:
			self[k] = kargs[k]

	def copy(self):
		return self.get_dict()


cdef class CollectionOperands(CollectionRef):
	#cdef object pylist

	def __init__(self, Collection c):
		if c.coll == NULL:
			raise RuntimeError("Uninitialized collection")
		self.set_collection(c.coll)
		self.init_pylist()

	cdef init_pylist(self):
		cdef XmmsValue v
		if self.pylist is None:
			v = XmmsValue()
			v.set_value(xmmsv_coll_operands_get(self.coll))
			self.pylist = v.value()

	def __repr__(self):
		return repr(self.pylist)
	def __str__(self):
		return str(self.pylist)
	def __len__(self):
		return len(self.pylist)
	def __getitem__(self, i):
		return self.pylist[i]
	def __delitem__(self, i):
		cdef Collection op
		try:
			op = self.pylist.pop(i)
			xmmsv_coll_remove_operand(self.coll, op.coll)
		except IndexError:
			raise IndexError("Index out of range")
	def __iter__(self):
		return iter(self.pylist)

	cpdef append(self, Collection op):
		"""Append an operand"""
		with cython.nonecheck(True):
			xmmsv_coll_add_operand(self.coll, op.coll)
			self.pylist.append(op)

	cpdef extend(self, ops):
		for o in ops:
			self.append(o)

	cpdef remove(self, Collection op):
		"""Remove an operand"""
		with cython.nonecheck(True):
			xmmsv_coll_remove_operand(self.coll, op.coll)
			self.pylist.remove(op)

	cpdef clear(self):
		cdef Collection op
		for op in self.pylist:
			xmmsv_coll_remove_operand(self.coll, op.coll)
		self.pylist = []

	cpdef list(self):
		return self.pylist[:]



cdef class CollectionIDList(CollectionRef):
	def __init__(self, Collection c):
		if c.coll == NULL:
			raise RuntimeError("Uninitialized collection")
		self.set_collection(c.coll)

	def __len__(self):
		return xmmsv_coll_idlist_get_size(self.coll)

	def __iter__(self):
		"""Iterate over ids"""
		cdef int x = -1
		cdef int l
		cdef int i = 0
		l = xmmsv_coll_idlist_get_size(self.coll)
		while i < l:
			if not xmmsv_coll_idlist_get_index(self.coll, i, &x):
				raise RuntimeError("Failed to retrieve id at index %d" % i)
			yield x
			i += 1

	def list(self):
		"""Returns a _COPY_ of the idlist as an ordinary list"""
		return list(self)

	def __repr__(self):
		return repr(self.list())

	cpdef append(self, int v):
		"""Appends an id to the idlist"""
		if not xmmsv_coll_idlist_append(self.coll, v):
			raise RuntimeError("Failed to append an id")

	cpdef extend(self, v):
		cdef int a
		for a in v:
			self.append(a)

	def __iadd__(self, v):
		self.extend(v)
		return self

	cpdef insert(self, int v, int i):
		"""Inserts an id at specified position"""
		if i < 0:
			i = len(self) + i
		if not xmmsv_coll_idlist_insert(self.coll, v, i):
			raise IndexError("Index out of range")

	cpdef remove(self, int i):
		"""Removes entry at specified position"""
		if i < 0:
			i = len(self) + i
		if not xmmsv_coll_idlist_remove(self.coll, i):
			raise IndexError("Index out of range")

	cpdef pop(self, int i = -1):
		"""Remove an id at specified position and return it"""
		x = self[i]
		self.remove(i)
		return x

	def __delitem__(self, int i):
		if i < 0:
			i = len(self) + i
		self.remove(i)

	def __getitem__(self, int i):
		cdef int x = 0
		if i < 0:
			i = len(self) + i
		if not xmmsv_coll_idlist_get_index(self.coll, i, &x):
			raise IndexError("Index out of range")
		return x

	def __setitem__(self, int i, int v):
		if i < 0:
			i = len(self) + i
		if not xmmsv_coll_idlist_set_index(self.coll, i, v):
			raise IndexError("Index out of range")

	cpdef clear(self):
		if not xmmsv_coll_idlist_clear(self.coll):
			raise RuntimeError("Failed to clear ids")


# XXX Might not be needed.
class CollectionWrapper(Collection):
	pass

class BaseCollection(CollectionWrapper):
	def __init__(Collection self, int _colltype, **kargs):
		self.coll = xmmsv_new_coll(<xmmsv_coll_type_t> _colltype)
		if self.coll == NULL:
			raise RuntimeError("Bad collection")

		operands = kargs.pop("operands", None)
		if operands:
			self.operands = operands

		ids = kargs.pop("ids", None)
		if ids:
			try:
				self.ids = ids
			except TypeError: # Just ignore idlist when illegal
				pass

		self.attributes.update(kargs)

class Reference(BaseCollection):
	def __init__(Collection self, ref = None, ns = "Collections", **kargs):
		kargs.update(
				namespace = kargs.get('namespace', ns),
				reference = kargs.get('reference', ref)
				)
		BaseCollection.__init__(self, XMMS_COLLECTION_TYPE_REFERENCE, **kargs)

class Universe(Reference):
	def __init__(self):
		Reference.__init__(self, reference = "All Media")

	def __repr__(self):
		return "%s()" % self.__class__.__name__

class FilterCollection(BaseCollection):
	def __init__(Collection self, _operation, parent = None, **kargs):
		if 'operands' in kargs:
			operands = kargs
		elif parent is None:
			operands = [Universe()]
		else:
			operands = [parent]

		BaseCollection.__init__(self, _operation,
				operands = operands,
				**kargs)

class Equals(FilterCollection):
	def __init__(Collection self, parent = None, **kargs):
		FilterCollection.__init__(self, XMMS_COLLECTION_TYPE_EQUALS, parent, **kargs)

class NotEqual(FilterCollection):
	def __init__(Collection self, parent = None, **kargs):
		FilterCollection.__init__(self, XMMS_COLLECTION_TYPE_NOTEQUAL, parent, **kargs)

class Smaller(FilterCollection):
	def __init__(Collection self, parent = None, **kargs):
		FilterCollection.__init__(self, XMMS_COLLECTION_TYPE_SMALLER, parent, **kargs)

class SmallerEqual(FilterCollection):
	def __init__(Collection self, parent = None, **kargs):
		FilterCollection.__init__(self, XMMS_COLLECTION_TYPE_SMALLEREQ, parent, **kargs)

class Greater(FilterCollection):
	def __init__(Collection self, parent = None, **kargs):
		FilterCollection.__init__(self, XMMS_COLLECTION_TYPE_GREATER, parent, **kargs)

class GreaterEqual(FilterCollection):
	def __init__(Collection self, parent = None, **kargs):
		FilterCollection.__init__(self, XMMS_COLLECTION_TYPE_GREATEREQ, parent, **kargs)

class Match(FilterCollection):
	def __init__(Collection self, parent = None, **kargs):
		FilterCollection.__init__(self, XMMS_COLLECTION_TYPE_MATCH, parent, **kargs)

class Token(FilterCollection):
	def __init__(Collection self, parent = None, **kargs):
		FilterCollection.__init__(self, XMMS_COLLECTION_TYPE_TOKEN, parent, **kargs)

class Has(FilterCollection):
	def __init__(Collection self, parent = None, **kargs):
		FilterCollection.__init__(self, XMMS_COLLECTION_TYPE_HAS, parent, **kargs)

class Order(BaseCollection):
	def __init__(Collection self, operand = None, field = None, direction = 0, **kargs):
		if isinstance(direction, int):
			kargs['direction'] = 'ASC' if direction >= 0 else 'DESC'
		else:
			direction = str(direction).upper()
			if direction not in ('ASC', 'DESC'):
				raise TypeError("'direction' must be an integer or one of 'ASC', 'DESC'")
			kargs['direction'] = direction

		kargs['operands'] = kargs.get('operands', [operand or Universe()])

		if field == 'id':
			kargs["type"] = "id"
		elif field:
			kargs["field"] = field
			kargs["type"] = "value"
		else:
			kargs["type"] = kargs.get('type', "random")

		if kargs["type"] == "random":
			del kargs['direction']

		BaseCollection.__init__(self, XMMS_COLLECTION_TYPE_ORDER, **kargs)

class Mediaset(BaseCollection):
	def __init__(Collection self, operand = None, **kargs):
		kargs['operands'] = kargs.get('operands', [operand or Universe()])
		BaseCollection.__init__(self, XMMS_COLLECTION_TYPE_MEDIASET, **kargs)

class Limit(BaseCollection):
	def __init__(Collection self, operand = None, start = 0, length = 0, **kargs):
		kargs.update(
				operands = kargs.get('operands', [operand or Universe()]),
				start = int(start),
				length = int(length)
				)
		BaseCollection.__init__(self, XMMS_COLLECTION_TYPE_LIMIT, **kargs)

class IDList(BaseCollection):
	def __init__(Collection self, ids = None, **kargs):
		kargs.update(
				type = kargs.get('type', 'list'),
				ids = ids
				)
		BaseCollection.__init__(self, XMMS_COLLECTION_TYPE_IDLIST, **kargs)

class Queue(IDList):
	def __init__(Collection self, ids = None, **kargs):
		kargs.update(
				type = kargs.get('type', 'queue')
				)
		IDList.__init__(self, ids, **kargs)

class PShuffle(IDList):
	def __init__(Collection self, ids = None, parent = None, **kargs):
		kargs.update(
				type = kargs.get('type', 'pshuffle'),
				operands = kargs.get('operands', [parent or Universe()])
				)
		IDList.__init__(self, ids, **kargs)

class Union(BaseCollection):
	def __init__(Collection self, *a, **kargs):
		kargs['operands'] = kargs.get('operands', list(a))
		BaseCollection.__init__(self, XMMS_COLLECTION_TYPE_UNION, **kargs)

class Intersection(BaseCollection):
	def __init__(Collection self, *a, **kargs):
		kargs['operands'] = kargs.get('operands', list(a))
		BaseCollection.__init__(self, XMMS_COLLECTION_TYPE_INTERSECTION, **kargs)

class Complement(BaseCollection):
	def __init__(Collection self, operand = None, **kargs):
		kargs['operands'] = kargs.get('operands', [operand or Universe()])
		BaseCollection.__init__(self, XMMS_COLLECTION_TYPE_COMPLEMENT, **kargs)

cdef _collclass = {
		XMMS_COLLECTION_TYPE_REFERENCE: Reference,
		XMMS_COLLECTION_TYPE_UNIVERSE: Universe,
		XMMS_COLLECTION_TYPE_UNION: Union,
		XMMS_COLLECTION_TYPE_INTERSECTION: Intersection,
		XMMS_COLLECTION_TYPE_COMPLEMENT: Complement,
		XMMS_COLLECTION_TYPE_HAS: Has,
		XMMS_COLLECTION_TYPE_EQUALS: Equals,
		XMMS_COLLECTION_TYPE_NOTEQUAL: NotEqual,
		XMMS_COLLECTION_TYPE_MATCH: Match,
		XMMS_COLLECTION_TYPE_TOKEN: Token,
		XMMS_COLLECTION_TYPE_SMALLER: Smaller,
		XMMS_COLLECTION_TYPE_SMALLEREQ: SmallerEqual,
		XMMS_COLLECTION_TYPE_GREATER: Greater,
		XMMS_COLLECTION_TYPE_GREATEREQ: GreaterEqual,
		XMMS_COLLECTION_TYPE_ORDER: Order,
		XMMS_COLLECTION_TYPE_MEDIASET: Mediaset,
		XMMS_COLLECTION_TYPE_LIMIT: Limit,
		XMMS_COLLECTION_TYPE_IDLIST: IDList,
		}
cdef create_coll(xmmsv_t *coll):
	cdef xmmsv_coll_type_t colltype
	cdef Collection c
	cdef const_char *idlist_type
	cdef object idtype_uni

	colltype = xmmsv_coll_get_type(coll)
	c = CollectionWrapper()
	if colltype in _collclass:
		c.__class__ = _collclass[colltype]
	else:
		raise RuntimeError("Unknown collection type")

	c.set_collection(coll)

	if colltype == XMMS_COLLECTION_TYPE_REFERENCE:
		if c.attributes.get("reference", "") == "All Media":
			c.__class__ = Universe
	elif colltype == XMMS_COLLECTION_TYPE_IDLIST:
		if xmmsv_coll_attribute_get_string(coll, "type", &idlist_type):
			idtype_uni = to_unicode(idlist_type)
			if idtype_uni == "queue":
				c.__class__ = Queue
			elif idtype_uni == "pshuffle":
				c.__class__ = PShuffle

	return c


def coll_parse(pattern):
	cdef xmmsv_t *coll = NULL

	_pattern = from_unicode(pattern)
	if not xmmsv_coll_parse(<char *>_pattern, &coll):
		raise ValueError("Unable to parse the pattern")
	return create_coll(coll)

cdef get_default_source_pref():
	cdef int i = 0
	sourcepref = []
	while xmmsv_default_source_pref[i] != NULL:
		sourcepref.append(to_unicode(xmmsv_default_source_pref[i]))
		i += 1
	return sourcepref



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import subprocess
import sys
import re
from collections import defaultdict

if len(sys.argv) != 2 and len(sys.argv) != 3:
    print("usage: %s <from> [<to>]" % sys.argv[0])
    raise SystemExit(1)

rev_start = sys.argv[1]

if len(sys.argv) > 2:
    rev_end = sys.argv[2]
else:
    rev_end = "HEAD"

command = ["git", "log", "--pretty=%H☃%an☃%s", rev_start + "..." + rev_end]

commits = defaultdict(lambda : defaultdict(list))
for line in subprocess.check_output(command).decode('utf-8').split("\n"):
    if not line:
        continue
    commit, author, comment = line.split("☃")
    if not (comment.startswith("BUG") or comment.startswith("FEATURE")):
        continue
    match = re.findall("(FEATURE|BUG)\(([0-9]+)\).+", comment)
    if len(match) != 1 and len(match[0]) != 2:
        print("skipping", match)

    _, bug_id = match[0]

    commits[author][int(bug_id)].append(commit)

for author in sorted(commits.keys()):
    print("'''%s'''" % author)
    for bugid in sorted(commits[author].keys()):
        print("* <mantis>%s</mantis>" % bugid)
        for commit in commits[author][bugid]:
            print("** <git norepo=1 repo='xmms2-devel.git'>%s</git>" % commit)



#!/usr/bin/env python
#  XMMS2 - X Music Multiplexer System
#  Copyright (C) 2003-2006 XMMS2 Team
# 
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#                   
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  This file is a part of the XMMS2 client tutorial #4
#  Let's try some lists and to show the playlist

import xmmsclient
import os
import sys

def get_mediainfo (xmms, id):
	"""
	This function is basically the same as
	tut3.py. But since we doing it
	repeatedly in the playlist getter, we
	move it to a separate function.

	print out artist, title and bitrate
	for each entry in the playlist.
	"""
	result = xmms.medialib_get_info(id)
	result.wait()
	if result.iserror():
		print "medialib get info returns error, %s" % result.get_error()
		sys.exit(1)

	minfo = result.value()
	try:
		artist = minfo["artist"]
	except KeyError:
		artist = "No artist"

	try:
		title = minfo["title"]
	except KeyError:
		title = "No title"

	try:
		bitrate = minfo["bitrate"]
	except KeyError:
		bitrate = 0

	print "artist = %s" % artist
	print "title = %s" % title
	print "bitrate = %i" % bitrate



"""
The first part of this program is
commented in tut1.py
"""
xmms = xmmsclient.XMMS("tutorial4")
try:
	xmms.connect(os.getenv("XMMS_PATH"))
except IOError, detail:
	print "Connection failed:", detail
	sys.exit(1)

"""
So let's look at lists. Lists returned from xmms2d
can only contain one type of values. So you either
have a list of strings, a list of ints or a list of
ints. In this case we ask for the whole current
playlist. It will return a result with a list of ints.
Each int is the id number of the entry.

The playlist has two important numbers: the entry
and the position. Each alteration command (move,
remove) works on the position of the entry rather
than the id. This is because you can have more
than one item of the same entry in the playlist.

first we ask for the playlist.
"""
result = xmms.playlist_list_entries()
result.wait()

if result.iserror():
	print "error when asking for the playlist, %s" % result.get_error()

"""
Now iterate the list.
"""
plist = result.value()
for id in plist:
	"""
	Now we have an id number saved in the id variable.
	Let's feed it to the function above (which
	is the same as we learned in tut3.py).
	and print out some pretty numbers.
	"""
	get_mediainfo(xmms, id)

sys.exit(0)



#!/usr/bin/env python
from subprocess import check_output
from operator import itemgetter
import os

def collect_hashes(*paths):
    hashes = []
    for path in paths:
        data = check_output("git --git-dir=%s/.git ls-tree -r HEAD" % path, shell=True).decode()
        for line in data.split("\n"):
            line = line.strip()
            if line:
                perms, typ, chash, fpath = line.split()
                hashes.append((perms, typ, chash, os.path.normpath(os.path.join(path, fpath))))

    result = []
    for (perms, typ, chash, path) in sorted(hashes, key=itemgetter(3)):
        result.append("%s %s %s\t %s" % (perms, typ, chash, path))

    return "\n".join(result)

print(check_output("git describe", shell=True).decode())
print(collect_hashes(".", "doc/tutorial"))
print(collect_hashes(".", "src/lib/s4"))



from xmmsvalue cimport Collection, XmmsValue
from cxmmsvalue cimport *
from cxmmsclient cimport *

cdef bint ResultNotifier(xmmsv_t *res, void *o)
cdef void ResultDestroyNotifier(void *o)

cdef class XmmsSourcePreference:
	cdef object sources

cdef class XmmsResult
cdef class XmmsResultTracker:
	cdef object results

	cdef track_result(self, XmmsResult r)
	cdef release_result(self, XmmsResult r)
	cdef disconnect_all(self, bint unset_result)

cdef class XmmsResult:
	cdef xmmsc_result_t *res
	cdef object _cb
	cdef bint _cb_issetup
	cdef XmmsSourcePreference source_pref
	cdef int ispropdict
	cdef XmmsResultTracker result_tracker

	cdef set_sourcepref(self, XmmsSourcePreference sourcepref)
	cdef set_result(self, xmmsc_result_t *res)
	cdef set_callback(self, XmmsResultTracker rt, cb)

	cpdef disconnect(self)
	cpdef wait(self)
	cpdef is_error(self)
	cpdef xmmsvalue(self)
	cpdef value(self)

ctypedef int VisResultCommand
cdef enum:
	VIS_RESULT_CMD_NONE = 0
	VIS_RESULT_CMD_INIT = 1
	VIS_RESULT_CMD_START = 2

cdef class XmmsVisResult(XmmsResult):
	cdef XmmsValue _val
	cdef VisResultCommand command
	cdef xmmsc_connection_t *conn

	cdef set_command(self, VisResultCommand cmd, xmmsc_connection_t *conn)
	cdef retrieve_error(self)
	cdef _init_xmmsvalue(self)
	cdef _start_xmmsvalue(self)
	cpdef xmmsvalue(self)

cdef class XmmsVisChunk:
	cdef short *data
	cdef int sample_count

	cdef set_data(self, short *data, int sample_count)
	cpdef get_buffer(self)
	cpdef get_data(self)

cdef class XmmsCore
cdef class XmmsServiceNamespace:
	cdef readonly XmmsCore xmms
	cdef readonly XmmsServiceNamespace parent
	cdef bint registered
	cdef object _bound_methods

	cpdef _walk(self, namespace=*, constant=*, method=*, broadcast=*, path=*, other=*, depth=*, deep=*)
	cpdef register_namespace(self, path, instance, infos)
	cpdef register_constant(self, path, instance, infos)
	cpdef register_method(self, path, instance, infos)
	cpdef register_broadcast(self, path, instance, infos)
	cpdef register(self)

cdef class service_broadcast:
	cdef public object name
	cdef public object doc
	cdef readonly XmmsServiceNamespace namespace
	cdef bind(self, XmmsServiceNamespace namespace, name=*)
	cpdef emit(self, value=*)

cdef xmmsv_t *service_method_proxy(xmmsv_t *pargs, xmmsv_t *nargs, void *udata)
cdef xmmsc_sc_namespace_t *_namespace_get(xmmsc_connection_t *c, path, bint create)

cdef class _XmmsServiceClient
cdef class client_broadcast:
	cdef XmmsCore _xmms
	cdef bint _async
	cdef readonly object _path
	cdef readonly int _clientid
	cdef readonly object name
	cdef readonly object docstring

cdef class client_method:
	cdef _XmmsServiceClient _parent
	cdef object _callback
	cdef readonly object _path
	cdef readonly int _clientid
	cdef readonly object name
	cdef readonly object docstring
	cdef readonly object inspect

cdef class _XmmsServiceClient:
	cdef readonly object _path
	cdef readonly bint _async
	cdef readonly XmmsCore _xmms
	cdef readonly int _clientid

cdef class XmmsProxy:
	cdef XmmsCore _get_xmms(self)

cdef class XmmsCore:
	cdef xmmsc_connection_t *conn
	cdef int isconnected
	cdef object disconnect_fun
	cdef object needout_fun
	cdef readonly XmmsSourcePreference source_preference
	cdef XmmsResultTracker result_tracker
	cdef readonly object clientname

	cdef new_connection(self)
	cpdef get_source_preference(self)
	cpdef set_source_preference(self, sources)
	cpdef _needout_cb(self, int i)
	cpdef _disconnect_cb(self)
	cpdef disconnect(self)
	cpdef ioin(self)
	cpdef ioout(self)
	cpdef want_ioout(self)
	cpdef set_need_out_fun(self, fun)
	cpdef get_fd(self)
	cpdef connect(self, path=*, disconnect_func=*)
	cdef XmmsResult _create_result(self, cb, xmmsc_result_t *res, Cls)
	cdef XmmsResult create_result(self, cb, xmmsc_result_t *res)
	cdef XmmsResult create_vis_result(self, cb, xmmsc_result_t *res, VisResultCommand cmd)

cdef void python_need_out_fun(int i, void *obj)
cdef void python_disconnect_fun(void *obj)
cpdef userconfdir_get()

cdef class XmmsApi(XmmsCore):
	cpdef int c2c_get_own_id(self)
	cpdef XmmsResult quit(self, cb=*)
	cpdef XmmsResult plugin_list(self, typ, cb=*)
	cpdef XmmsResult playback_start(self, cb=*)
	cpdef XmmsResult playback_stop(self, cb=*)
	cpdef XmmsResult playback_tickle(self, cb=*)
	cpdef XmmsResult playback_pause(self, cb=*)
	cpdef XmmsResult playback_current_id(self, cb=*)
	cpdef XmmsResult playback_seek_ms(self, int ms, xmms_playback_seek_mode_t whence=*, cb=*)
	cpdef XmmsResult playback_seek_samples(self, int samples, xmms_playback_seek_mode_t whence=*, cb=*)
	cpdef XmmsResult playback_status(self, cb=*)
	cpdef XmmsResult broadcast_playback_status(self, cb=*)
	cpdef XmmsResult broadcast_playback_current_id(self, cb=*)
	cpdef XmmsResult playback_playtime(self, cb=*)
	cpdef XmmsResult signal_playback_playtime(self, cb=*)
	cpdef XmmsResult playback_volume_set(self, channel, int volume, cb=*)
	cpdef XmmsResult playback_volume_get(self, cb=*)
	cpdef XmmsResult broadcast_playback_volume_changed(self, cb=*)
	cpdef XmmsResult broadcast_playlist_loaded(self, cb=*)
	cpdef XmmsResult playlist_load(self, playlist, cb=*)
	cpdef XmmsResult playlist_list(self, cb=*)
	cpdef XmmsResult playlist_remove(self, playlist, cb=*)
	cpdef XmmsResult playlist_shuffle(self, playlist=*, cb=*)
	cpdef XmmsResult playlist_rinsert(self, int pos, url, playlist=*, cb=*, encoded=*)
	cpdef XmmsResult playlist_insert_url(self, int pos, url, playlist=*, cb=*, encoded=*)
	cpdef XmmsResult playlist_insert_id(self, int pos, int id, playlist=*, cb=*)
	cpdef XmmsResult playlist_insert_collection(self, int pos, Collection coll, order=*, playlist=*, cb=*)
	cpdef XmmsResult playlist_radd(self, url, playlist=*, cb=*, encoded=*)
	cpdef XmmsResult playlist_add_url(self, url, playlist=*, cb=*, encoded=*)
	cpdef XmmsResult playlist_add_id(self, int id, playlist=*, cb=*)
	cpdef XmmsResult playlist_add_collection(self, Collection coll, order=*, playlist=*, cb=*)
	cpdef XmmsResult playlist_remove_entry(self, int id, playlist=*, cb=*)
	cpdef XmmsResult playlist_clear(self, playlist=*, cb=*)
	cpdef XmmsResult playlist_list_entries(self, playlist=*, cb=*)
	cpdef XmmsResult playlist_sort(self, props, playlist=*, cb=*)
	cpdef XmmsResult playlist_set_next_rel(self, int position, cb=*)
	cpdef XmmsResult playlist_set_next(self, int position, cb=*)
	cpdef XmmsResult playlist_move(self, int cur_pos, int new_pos, playlist=*, cb=*)
	cpdef XmmsResult playlist_create(self, playlist, cb=*)
	cpdef XmmsResult playlist_current_pos(self, playlist=*, cb=*)
	cpdef XmmsResult playlist_current_active(self, cb=*)
	cpdef XmmsResult broadcast_playlist_current_pos(self, cb=*)
	cpdef XmmsResult broadcast_playlist_changed(self, cb=*)
	cpdef XmmsResult broadcast_config_value_changed(self, cb=*)
	cpdef XmmsResult config_set_value(self, key, val, cb=*)
	cpdef XmmsResult config_get_value(self, key, cb=*)
	cpdef XmmsResult config_list_values(self, cb=*)
	cpdef XmmsResult config_register_value(self, valuename, defaultvalue, cb=*)
	cpdef XmmsResult medialib_add_entry(self, path, cb=*, encoded=*)
	cpdef XmmsResult medialib_remove_entry(self, int id, cb=*)
	cpdef XmmsResult medialib_move_entry(self, int id, url, cb=*, encoded=*)
	cpdef XmmsResult medialib_get_info(self, int id, cb=*)
	cpdef XmmsResult medialib_rehash(self, int id=*, cb=*)
	cpdef XmmsResult medialib_get_id(self, url, cb=*, encoded=*)
	cpdef XmmsResult medialib_import_path(self, path, cb=*, encoded=*)
	cpdef XmmsResult medialib_property_set(self, int id, key, value, source=*, cb=*)
	cpdef XmmsResult medialib_property_remove(self, int id, key, source=*, cb=*)
	cpdef XmmsResult broadcast_medialib_entry_added(self, cb=*)
	cpdef XmmsResult broadcast_medialib_entry_updated(self, cb=*)
	cpdef XmmsResult broadcast_medialib_entry_removed(self, cb=*)
	cpdef XmmsResult broadcast_collection_changed(self, cb=*)
	cpdef XmmsResult signal_mediainfo_reader_unindexed(self, cb=*)
	cpdef XmmsResult broadcast_mediainfo_reader_status(self, cb=*)
	cpdef XmmsResult xform_media_browse(self, url, cb=*, encoded=*)
	cpdef XmmsResult coll_get(self, name, ns=*, cb=*)
	cpdef XmmsResult coll_list(self, ns=*, cb=*)
	cpdef XmmsResult coll_save(self, Collection coll, name, ns=*, cb=*)
	cpdef XmmsResult coll_remove(self, name, ns=*, cb=*)
	cpdef XmmsResult coll_rename(self, oldname, newname, ns=*, cb=*)
	cpdef XmmsResult coll_idlist_from_playlist_file(self, path, cb=*)
	cpdef XmmsResult coll_query(self, Collection coll, fetch, cb=*)
	cpdef XmmsResult coll_query_ids(self, Collection coll, start=*, leng=*, order=*, cb=*)
	cpdef XmmsResult coll_query_infos(self, Collection coll, fields, start=*, leng=*, order=*, groupby=*, cb=*)
	#C2C
	#Low level API (internal ?) that should not be used directly.
	#cdef xmmsc_c2c_reply_policy_t get_c2c_policy(self, reply_policy=*)
	#cpdef XmmsResult c2c_send(self, int dest, payload, reply_policy=*, cb=*)
	#cpdef XmmsResult c2c_reply(self, int msgid, payload, reply_policy=*, cb=*)
	#cpdef XmmsResult broadcast_c2c_message(self, cb=*)
	cpdef XmmsResult c2c_ready(self, cb=*)
	cpdef XmmsResult c2c_get_connected_clients(self, cb=*)
	cpdef XmmsResult c2c_get_ready_clients(self, cb=*)
	cpdef XmmsResult broadcast_c2c_ready(self, cb=*)
	cpdef XmmsResult broadcast_c2c_client_connected(self, cb=*)
	cpdef XmmsResult broadcast_c2c_client_disconnected(self, cb=*)
	#Services
	cpdef bint sc_init(self)
	cpdef bint sc_broadcast_emit(self, broadcast, value=*)
	cpdef XmmsResult sc_broadcast_subscribe(self, int dest, broadcast, cb=*)
	cpdef XmmsResult sc_call(self, int dest, method, args=*, kargs=*, cb=*)
	cpdef XmmsResult sc_introspect_namespace(self, int dest, path=*, cb=*)
	cpdef XmmsResult sc_introspect_method(self, int dest, path, cb=*)
	cpdef XmmsResult sc_introspect_broadcast(self, int dest, path, cb=*)
	cpdef XmmsResult sc_introspect_constant(self, int dest, path, cb=*)
	cpdef XmmsResult sc_introspect_docstring(self, int dest, path, cb=*)
	#
	cpdef XmmsResult bindata_add(self, data, cb=*)
	cpdef XmmsResult bindata_retrieve(self, hash, cb=*)
	cpdef XmmsResult bindata_remove(self, hash, cb=*)
	cpdef XmmsResult bindata_list(self, cb=*)
	cpdef XmmsResult stats(self, cb=*)
	cpdef XmmsResult visualization_version(self, cb=*)
	cpdef XmmsResult visualization_init(self, cb=*)
	cpdef XmmsResult visualization_start(self, int handle, cb=*)
	cpdef bint visualization_started(self, int handle)
	cpdef bint visualization_errored(self, int handle)
	cpdef XmmsResult visualization_property_set(self, int handle, key, value, cb=*)
	cpdef XmmsResult visualization_properties_set(self, int handle, props=*, cb=*)
	cpdef XmmsVisChunk visualization_chunk_get(self, int handle, int drawtime=*, bint blocking=*)
	cpdef visualization_shutdown(self, int handle)

cdef class XmmsLoop(XmmsApi):
	cdef bint do_loop
	cdef object wakeup



class compiler_flags(object):
    def __init__(self, conf):
        self.conf = conf
        self.extra = { "CFLAGS": [], "CXXFLAGS": [] }
        if self.conf.check_cc(cflags="-Werror=unknown-warning-option", mandatory=False):
            self.extra["CFLAGS"] = ["-Werror=unknown-warning-option"]
        if self.conf.check_cxx(cxxflags="-Werror=unknown-warning-option", mandatory=False):
            self.extra["CXXFLAGS"] = ["-Werror=unknown-warning-option"]

        langs = (
            (self.conf.check_cc, "CFLAGS", "c"),
            (self.conf.check_cxx, "CXXFLAGS", "cxx")
        )

        kinds = (
            ("feature", "-f", "-fno-"),
            ("warning", "-W", "-Wno-"),
            ("error", "-Werror=", "-Wno-error=")
        )

        # Generate functions like flags.enable_cxx_warning(flag_name)
        for conf_func, flag, lang_name in langs:
            for kind, enable_prefix, disable_prefix in kinds:
                for name, func in self._generate_funcs(conf_func, flag, enable_prefix, disable_prefix):
                    self.__dict__["%s_%s_%s" % (name, lang_name, kind)] = func

        # Generate functions like flags.enable_warning(flag_name) which calls both c and cxx.
        for action in ("check", "enable", "disable"):
            for kind, _, _ in kinds:
                self.__dict__["%s_%s" % (action, kind)] = self._generate_multi_func(action, kind, [lang for _, _, lang in langs])

    def _generate_funcs(self, func, flag, enable_prefix, disable_prefix):
        def check(flag_name):
            return self._check(func, flag, enable_prefix, disable_prefix, flag_name)

        def enable(flag_name):
            enable, disable = check(flag_name)
            if enable:
                self.conf.env.append_unique(flag, [enable])

        def disable(flag_name):
            enable, disable = check(flag_name)
            if disable:
                self.conf.env.append_unique(flag, [disable])

        return (("check", check), ("enable", enable), ("disable", disable))

    def _generate_multi_func(self, action, kind, langs):
        def multi(flag_name):
            for lang in langs:
                self.__dict__["%s_%s_%s" % (action, lang, kind)](flag_name)
        return multi

    def _check(self, func, key, enable_prefix, disable_prefix, name):
        enable_flag = enable_prefix + name
        disable_flag = disable_prefix + name

        # Fork the environment so we can add some temporary compiler
        # flags that might be needed during the detection.
        self.conf.env.stash()
        self.conf.env[key] += self.extra[key]

        kw = { 'mandatory': False, key.lower(): enable_flag }
        result = func(**kw)

        self.conf.env.revert()

        if not result:
            return False, False

        uselib = name.replace("-", "").replace("=", "").upper()
        self.conf.env.append_unique("%s_ENABLE_%s" % (key, uselib), [enable_flag])
        self.conf.env.append_unique("%s_DISABLE_%s" % (key, uselib), [disable_flag])

        return enable_flag, disable_flag



#!/usr/bin/env python
import sys
import wave
import struct
import math
from itertools import count, repeat
from cStringIO import StringIO
from contextlib import closing
from subprocess import call

CHANNELS = 1
FORMAT = 16
SAMPLERATE = 44100
DURATION = 10

notes = {
    "C4": 261.63,
    "D4": 293.66,
    "E4": 329.63,
    "F4": 349.23,
    "G4": 392.00,
    "A4": 440.00,
    "B4": 493.88
}

def sine_wave(frequency=440.0, samplerate=44100, amplitude=0.5, repeat=0):
    """
    Slightly modified version of : http://zacharydenton.com/generate-audio-with-python/
    """
    period = int(samplerate / frequency)
    if amplitude > 1.0: amplitude = 1.0
    if amplitude < 0.0: amplitude = 0.0
    lookup_table = [float(amplitude) * math.sin(2.0*math.pi*float(frequency)*(float(i%period)/float(samplerate))) for i in xrange(period)]
    return (lookup_table[i%period]*samplerate for i in xrange(repeat))

def write(filename, freq):
    print "* Writing %(duration)ds of %(freq).2fHz to %(filename)s" % { "duration": DURATION, "freq": freq, "filename": filename }
    with closing(wave.open(filename, "wb")) as fd:
        fd.setparams((CHANNELS, FORMAT / 8, SAMPLERATE, SAMPLERATE * DURATION, 'NONE', 'noncompressed'))
        buffer = StringIO()
        for x in sine_wave(freq, repeat=SAMPLERATE * DURATION):
            buffer.write(struct.pack("h" * CHANNELS, *([x] * CHANNELS)))
        fd.writeframes(buffer.getvalue())

def ffmpeg(name, params, input, output, tags={}):
    print "* Encoding %(name)s from %(input)s to %(output)s" % { "name": name, "input": input, "output": output }
    metadata = []
    for tag, value in tags.items():
        metadata += ["-metadata", tag + "=" + value]
    if call(["ffmpeg", "-y", "-loglevel", "error", "-i", input] + params + metadata + [output]) != 0:
        print "x FAILED!"

write("test.wav", 440)

metadata = { "artist": "Professor Hertz", "title": "Aaaa yeah" }

ffmpeg("MP4+AAC", ["-ac", "2", "-c:a", "libfaac"],
       "test.wav", "test.m4a", metadata)
ffmpeg("WavPack", ["-ac", "2", "-c:a", "wavpack"],
       "test.wav", "test.wv", metadata)
ffmpeg("AC-3", ["-ac", "2", "-c:a", "ac3"],
       "test.wav", "test.ac3", metadata)
ffmpeg("DTS", ["-ac", "2", "-strict", "experimental", "-c:a", "dts"],
       "test.wav", "test.dts", metadata)
ffmpeg("FLAC", ["-ac", "2", "-c:a", "flac"],
       "test.wav", "test.flac", metadata)
ffmpeg("MP3", ["-ac", "2", "-c:a", "mp3"],
       "test.wav", "test.mp3", metadata)
ffmpeg("FLV+NellyMoser", ["-ac", "1", "-c:a", "nellymoser"],
       "test.wav", "test+nellymoser.flv", metadata)
ffmpeg("FLV+AAC", ["-ac", "2", "-c:a", "libfaac"],
       "test.wav", "test+aac.flv", metadata)
ffmpeg("FLV+MP3", ["-ac", "2", "-c:a", "mp3"],
       "test.wav", "test+mp3.flv", metadata)
ffmpeg("FLV+PCM", ["-ac", "2", "-c:a", "pcm_s16le"],
       "test.wav", "test+pcm.flv", metadata)
ffmpeg("FLV+SWF", ["-ac", "2", "-c:a", "adpcm_swf"],
       "test.wav", "test+swf.flv", metadata)
ffmpeg("Ogg+Opus", ["-ac", "2", "-c:a", "opus"],
       "test.wav", "test.opus", metadata)
ffmpeg("Ogg+Vorbis", ["-ac", "2", "-c:a", "libvorbis"],
       "test.wav", "test.ogg", metadata)
ffmpeg("Ogg+Speex", ["-ac", "2", "-c:a", "libspeex"],
       "test.wav", "test.spx", metadata)
ffmpeg("WMA v1", ["-ac", "1", "-c:a", "wmav1"],
       "test.wav", "test+wmav1.wma", metadata)
ffmpeg("WMA v2", ["-ac", "1", "-c:a", "wmav2"],
       "test.wav", "test+wmav2.wma", metadata)
ffmpeg("AIFF", ["-ac", "2", "-c:a", "pcm_s16be"],
       "test.wav", "test.aiff", metadata)
ffmpeg("CAF", ["-ac", "2", "-c:a", "pcm_s16be"],
       "test.wav", "test.caf", metadata)
ffmpeg("AU", ["-ac", "2", "-c:a", "pcm_s16be"],
       "test.wav", "test.au", metadata)

# No TTA muxer in ffmpeg at this point in time.
#ffmpeg("TTA", ["-f", "rawvideo", "-ac", "2", "-c:a", "ttaenc"],
#       "test.wav", "test.tta", metadata)



import sys

class Indenter:
	indent = 0

	@classmethod
	def enter(cls, s = None):
		if s != None:
			Indenter.printline(s)

		cls.indent += 1

	@classmethod
	def leave(cls, s = None):
		cls.indent -= 1

		Indenter.printline(s)

	# print the given string without a trailing newline
	@classmethod
	def printx(cls, s):
		sys.stdout.write('\t' * cls.indent)
		sys.stdout.write(s)

	# print the given string including a trailing newline
	@classmethod
	def printline(cls, s = ''):
		if s:
			cls.printx(s)
		print('')



#!/usr/bin/env python
import sys
import string
import xml.dom.minidom
from collections import OrderedDict

def _attribute_value(xml_element, attribute, default=None):
	try:
		return xml_element.attributes[attribute].value
	except KeyError:
		return default

class Enum(OrderedDict):
	def __init__(self, name, namespace_hint=None):
		OrderedDict.__init__(self)

		self.name = name
		if namespace_hint:
			self.namespace_hint = namespace_hint
		else:
			self.namespace_hint = self.name.upper()

	def add_member(self, name, value=None, alias=None):
		name = name.upper()
		if name in self:
			raise KeyError("%s.%s already defined" % (self.name, name))
		m = EnumMember(name, self, value, alias)
		self[name] = m
		return m

	def __setitem__(self, k, member):
		if not isinstance(member, EnumMember):
			raise ValueError("Expected EnumMember, got %s" % type(member).__name__)
		return OrderedDict.__setitem__(self, k, member)

class EnumMember:
	def __init__(self, name, enum, value=None, alias=None):
		self.name = name
		self.enum = enum
		if alias is not None and value is not None:
			raise ValueError("value and alias are mutually exclusive")
		if value is not None:
			self.value = int(value)
		elif alias:
			self.alias = alias

	def fullname(self, sep='_', fmt=str):
		return '%s%s%s' % (self.enum.namespace_hint, sep, self.name)

	def dottedname(self):
		return '%s.%s' % (self.enum.name, self.name)

	def __repr__(self):
		value = getattr(self, 'value', None)
		rval = ''
		if value is None:
			alias = getattr(self, 'alias', None)
			if alias is not None:
				val, src, t = alias
				rval = '%s%s%s%s' % (src or '', src and '.' or '', val, t == 'constant' and ' (constant)')
		else:
			rval = '%d' % value

		return "<%s %s%s%s>" % (type(self).__name__, self.dottedname(), rval and '=', rval)

class Constant:
	def __init__(self, name, value):
		self.name = name
		self.value = value

class NamedElement:
	def __init__(self, xml_element):
		name_element = xml_element.getElementsByTagName('name')[0]
		self.name = name_element.firstChild.data.strip()

class DocumentedElement:
	def __init__(self, xml_element):
		doc_element = xml_element.getElementsByTagName('documentation')[0]
		self.documentation = doc_element.firstChild.data.strip()

class TypedElement:
	def __init__(self, xml_element):
		self.type = []

		type_element = xml_element.getElementsByTagName('type')[0]

		choices = [e for e in type_element.childNodes
		           if e.nodeType == e.ELEMENT_NODE]
		node = choices[0]

		while node:
			nodeinfo = node.nodeName
			if nodeinfo == 'enum-value':
				nodeinfo = (nodeinfo, node.attributes['name'].value)
			self.type.append(nodeinfo)

			choices = [e for e in node.childNodes
			           if e.nodeType == e.ELEMENT_NODE]
			node = (choices or [None])[0]

class IpcFoo:
	def __init__(self, xml_element):
		self.version = int(xml_element.attributes.get('version').value)
		self.objects = []

		object_elements = xml_element.getElementsByTagName('object')
		object_id = 1 # ID 0 is reserved for signal voodoo

		self.constants = OrderedDict()

		ipc_version = Constant('IPC_PROTOCOL_VERSION', self.version)
		self.constants[ipc_version.name] = ipc_version

		for c in xml_element.getElementsByTagName('constant'):
			if c.parentNode is not xml_element:
				continue
			const = IpcConstant(c)
			self.constants[const.name] = const

		self.enums = OrderedDict()

		# Generated enums
		obj_enum = self.objects_enum = Enum('ipc_object', 'IPC_OBJECT')
		sig_enum = self.signals_enum = Enum('ipc_signal', 'IPC_SIGNAL')
		self.enums[obj_enum.name] = obj_enum
		self.enums[sig_enum.name] = sig_enum

		obj_enum.add_member('SIGNAL') # special object for signals and broadcasts

		for e in xml_element.getElementsByTagName('enum'):
			if e.parentNode is not xml_element:
				continue
			enum = IpcEnum(e)
			self.enums[enum.name] = enum

		for object_element in object_elements:
			object = IpcObject(object_element)
			self.objects.append(object)

			m = obj_enum.add_member(object.name.upper())
			object.id = m

			methods_enum = Enum('ipc_command_%s' % object.name)
			self.enums[methods_enum.name] = methods_enum
			alias_first = ('IPC_COMMAND_FIRST', None, 'constant')
			for meth in object.methods:
				m = methods_enum.add_member(meth.name.upper(), alias=alias_first)
				meth.id = m
				alias_first = None

		obj_enum.add_member('END')

		for bcsig, obj, type in self.iter_broadcasts_and_signals():
			m = sig_enum.add_member('%s_%s' % (obj.name.upper(), bcsig.name.upper()))
			bcsig.id = m
		sig_enum.add_member('END')


	def iter_broadcasts_and_signals(self):
		for obj in self.objects:
			for bc in obj.broadcasts:
				yield (bc, obj, 'broadcast')
			for s in obj.signals:
				yield (s, obj, 'signal')

class IpcConstant(Constant):
	_value_parse = {
			'string': str,
			'integer': int,
			'float': float,
			}
	def __init__(self, xml_element):
		name = NamedElement(xml_element).name

		value_element = xml_element.getElementsByTagName('value')[0]
		value_type = _attribute_value(value_element, 'type', 'string')
		value = value_element.firstChild.data.strip()
		value = self._value_parse[value_type](value)

		Constant.__init__(self, name, value)

class IpcEnum(Enum):
	def __init__(self, xml_element):
		name = NamedElement(xml_element).name
		ns_hint = xml_element.getElementsByTagName('namespace-hint')
		if ns_hint:
			ns_hint = ns_hint[0].firstChild.data.strip().upper()
		else: 
			ns_hint = None
		Enum.__init__(self, name, ns_hint)

		for m in xml_element.getElementsByTagName('member'):
			member = IpcEnumMember(m, self)
			name = member.name
			if name in self:
				raise KeyError("Enum member %s.%s defined twice" % (self.name, name))
			self[name] = member

class IpcEnumMember(EnumMember):
	def __init__(self, xml_element, enum):
		name = xml_element.firstChild.data.strip().upper()
		value = _attribute_value(xml_element, 'value', None)
		ref_type = _attribute_value(xml_element, 'ref-type', 'enum')
		ref_source = _attribute_value(xml_element, 'ref-source', None)
		ref_value = _attribute_value(xml_element, 'ref-value', None)
		if ref_value:
			alias = (ref_value, ref_source or None, ref_type or 'enum')
		else:
			alias = None
		EnumMember.__init__(self, name, enum, value, alias)

class IpcObject(NamedElement):
	def __init__(self, xml_element):
		NamedElement.__init__(self, xml_element)

		self.id = 0
		self.methods = []
		self.broadcasts = []
		self.signals = []

		method_elements = xml_element.getElementsByTagName('method')
		#method_id = 32 # IDs 0..31 are reserved for voodoo use

		for method_element in method_elements:
			method = IpcMethod(method_element)

			#method.id = method_id
			#method_id += 1

			self.methods.append(method)

		signal_elements = xml_element.getElementsByTagName('signal')

		for signal_element in signal_elements:
			self.signals.append(IpcSignalOrBroadcast(signal_element))

		broadcast_elements = xml_element.getElementsByTagName('broadcast')

		for broadcast_element in broadcast_elements:
			self.broadcasts.append(IpcSignalOrBroadcast(broadcast_element))

class IpcMethod(NamedElement, DocumentedElement):
	def __init__(self, xml_element):
		NamedElement.__init__(self, xml_element)
		DocumentedElement.__init__(self, xml_element)

		self.id = 0
		self.arguments = []
		self.return_value = None

		# Store attributes for this method as boolean object attributes
		for attr in ('noreply', 'need_cookie', 'need_client'):
			val = xml_element.attributes.get(attr, False)
			if val:
				val = (val.value == "true")
			setattr(self, attr, val)

		argument_elements = xml_element.getElementsByTagName('argument')

		for argument_element in argument_elements:
			self.arguments.append(IpcMethodArgument(argument_element))

		return_value_elements = xml_element.getElementsByTagName('return_value')
		if return_value_elements:
			self.return_value = IpcReturnValue(return_value_elements[0])

class IpcMethodArgument(NamedElement, DocumentedElement, TypedElement):
	def __init__(self, xml_element):
		NamedElement.__init__(self, xml_element)
		DocumentedElement.__init__(self, xml_element)
		TypedElement.__init__(self, xml_element)
		default_elements = xml_element.getElementsByTagName('default-hint')
		if default_elements:
			self.default_hint = default_elements[0].firstChild.data.strip()

class IpcReturnValue(DocumentedElement, TypedElement):
	def __init__(self, xml_element):
		DocumentedElement.__init__(self, xml_element)
		TypedElement.__init__(self, xml_element)

class IpcSignalOrBroadcast(NamedElement, DocumentedElement):
	def __init__(self, xml_element):
		NamedElement.__init__(self, xml_element)
		DocumentedElement.__init__(self, xml_element)

		self.id = None
		self.return_value = None

		#id_element = xml_element.getElementsByTagName('id')[0]
		#self.id = int(id_element.firstChild.data.strip())

		return_value_elements = xml_element.getElementsByTagName('return_value')
		self.return_value = IpcReturnValue(return_value_elements[0])

def parse_xml(file):
    #load the xml file
    doc = xml.dom.minidom.parse(file)

    ipc_element = doc.getElementsByTagName('ipc')[0]

    return IpcFoo(ipc_element)



from cpython.unicode cimport PyUnicode_DecodeUTF8, PyUnicode_AsUTF8String
from libc.string cimport const_char, strlen

cdef inline to_unicode(const_char *s):
	return PyUnicode_DecodeUTF8(s, strlen(s), "replace")

cdef inline from_unicode(object o):
	if isinstance(o, unicode):
		return PyUnicode_AsUTF8String(o)
	else:
		return o



#!/usr/bin/python3.4
import os
import sys
import re
import hashlib
import gzip
import pprint
import math

dir_pattern = re.compile("^(?!(\.conf_check|c4che))")
file_pattern = re.compile("^(?!(\.(wafpickle|lock-waf).*|.+\.(log|a|o)$))")

def open_any(path):
    if path.endswith(".gz"):
        return gzip.open(path, "rb")
    return open(path, "rb")

def analyze(payload):
    return (hashlib.sha1(payload).hexdigest(), len(payload))

def checksum_root(root):
    root = root.rstrip("/")
    checksums = {}
    for path, directories, files in os.walk(root, topdown=True):
        if path == root:
            directories[:] = filter(dir_pattern.match, directories)
        for name in filter(file_pattern.match, files):
            realpath = os.path.join(path, name)
            if os.path.islink(realpath):
                continue
            with open_any(realpath) as fd:
                checksums[realpath[len(root) + 1:]] = analyze(fd.read())
    return checksums

def compare(root_a, root_b, verbose=False):
    a = checksum_root(root_a)
    b = checksum_root(root_b)

    max_length = len(max(a.keys(), key=len))
    name_fmt = "%%(entry)-%ds" % max_length
    union = set(a.keys()) | set(b.keys())
    diffs = len(union)
    for name in sorted(union):
        checksum_a, size_a = a.get(name, (None, 0))
        checksum_b, size_b = b.get(name, (None, 0))

        args = {
            "entry": name, "root_a": root_a, "root_b": root_b,
            "sign": "=><"[(size_a > size_b) - (size_a < size_b)],
            "delta": math.fabs(size_a - size_b)
        }

        if checksum_a == checksum_b:
            if verbose:
                print((name_fmt + " success") % args)
            diffs -= 1
        elif not checksum_a:
            print((name_fmt + " missing in %(root_a)s") % args)
        elif not checksum_b:
            print((name_fmt + " missing in %(root_b)s") % args)
        else:
            print((name_fmt + " size(%(root_a)s %(sign)s %(root_b)s) %(delta)8d bytes") % args)

    print("Total: %d of %d failed" % (diffs, len(union)))

if __name__ == "__main__":
    compare(sys.argv[-2], sys.argv[-1], "-v" in sys.argv)



#!/usr/bin/env python
from subprocess import check_output
import os

authorchanges = {}
prevrelease = "HEAD"
prevdate = "Not released"
prevtreehash = "HEAD"

for line in check_output("git log --pretty=format:'%t\t%an\t%ai\t%s'", shell=True).decode('utf-8', "replace").split("\n"):
    line = line.strip()
    if not line:
        continue

    parts = line.strip().split("\t", 3)
    if len(parts) == 4:
        treehash, author, date, subject = parts
    else:
        treehash, author, date, subject = parts + ["No Subject"]

    if subject.lower().startswith("merge"):
        continue

    if subject.lower().startswith("release"):
        subject = subject[9:]
        if len(authorchanges) == 0:
            prevrelease = subject
            prevtreehash = treehash
            prevdate = date.split()[0]
            continue
        print("Changes between %s and %s" % (subject, prevrelease))
        print
        print(" Release date: %s" % prevdate)
        print(" Authors contributing to this release: %d" % len(authorchanges))
        print(" Number of changesets: %d" % sum(map(len, authorchanges.values())))
        print(" Number of files in this release: %s" % check_output("git ls-tree -r %s | wc -l" % prevtreehash, shell=True).decode().strip())
        print
        authors = list(authorchanges.keys())
        authors.sort()
        for a in authors:
            print(" %s:" % a)
            changes = authorchanges[a]
            changes.sort()
            for c in changes:
                print("  * %s" % c)
            print
        print
        print
        authorchanges={}
        prevrelease = subject
        prevtreehash = treehash
        prevdate = date.split()[0]
        continue

    if author not in authorchanges:
        authorchanges[author] = []

    authorchanges[author].append(subject)
